#include <stdio.h>
#include <math.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf_gamma.h>

#include "availability.h"

#define EPSILON 0.00000001

static int buildTraceFromChainMarkov(int seed, long int length, 
           markov_chain_t *chain, availability_trace_t *trace);
static int buildTraceFromChainPareto(int seed, long int length, double
           pareto_shape, markov_chain_t *chain, availability_trace_t *trace);
static int buildTraceFromChainWeibull(int seed, long int length, double
           weibull_shape, markov_chain_t *chain, availability_trace_t *trace);

/* computePiDistribution(): for a 3x3 chain */
int computePiDistribution(markov_chain_t *chain)
{

  double uu, ur, ud;
  double ru, rr, rd;
  double du, dr, dd;

  double *pu = &(chain->pi[0]);
  double *pr = &(chain->pi[1]);
  double *pd = &(chain->pi[2]);

  uu = chain->matrix[0][0];
  ur = chain->matrix[0][1];
  ud = chain->matrix[0][2];
  ru = chain->matrix[1][0];
  rr = chain->matrix[1][1];
  rd = chain->matrix[1][2];
  du = chain->matrix[2][0];
  dr = chain->matrix[2][1];
  dd = chain->matrix[2][2];

  if (fabs(ru - du) < EPSILON) {
    *pu = du / (1 - uu + du);
    *pr = (-dr -(ur - dr)*(*pu)) / (rr -1 - dr);
    *pd = 1 - *pu - *pr;
  } else if (fabs(ur - dr) < EPSILON) {
    *pr = dr / (1 - rr + dr);
    *pu = (-du -(ru - du)*(*pr)) / (uu -1 - du);
    *pd = 1 - *pu - *pr;
  } else {
    *pr = (-du + dr *(uu -1 -du)/(ur-dr)) / (ru - du - (rr - 1 -dr)*(uu -1 -du)/(ur - dr));
    *pu = (-dr - (rr -1 - dr) * (*pr)) / (ur - dr);
    *pd = 1 - *pr - *pu;
  }
  return 0;
}

/* pickAmongThree() */
int pickAmongThree(double *p, gsl_rng *rng)
{
  double ran;

  if (fabs(p[0]+p[1]+p[2] - 1) > EPSILON) {
    fprintf(stderr,"pickAmongThree(): probabilities don't add to 1\n");
    return -1;
  }

  ran=gsl_rng_uniform(rng);

  if (ran <= p[0]) {
    return 0;
  } else if (ran <= p[0]+p[1]) {
    return 1;
  } else {
    return 2;
  }
}

/* buildTraceFromChain() */
int buildTraceFromChain(int seed, long int length, 
                        distribution_t dist, double dist_shape, 
                        markov_chain_t *chain, availability_trace_t *trace)
{
  if (dist == MARKOV) {
    return buildTraceFromChainMarkov(seed,length,chain,trace);
  } else if (dist == PARETO) {
    return buildTraceFromChainPareto(seed,length,dist_shape,chain,trace);
  } else if (dist == WEIBULL) {
    return buildTraceFromChainWeibull(seed,length,dist_shape,chain,trace);
  } else {
    return -1;
  }
}

/* buildTraceFromChainNoPareto() */
static int buildTraceFromChainMarkov(int seed, long int length, 
           markov_chain_t *chain, availability_trace_t *trace)
{
  gsl_rng *rng;
  int new_state, current_state;
  int time_step;

  /* Initialize random number generator */
  rng = gsl_rng_alloc(gsl_rng_taus2);
  gsl_rng_set(rng, seed);

  /* Initialize the trace */
  trace->num_entries = 0;
  trace->dates = NULL;
  trace->states = NULL;

  /* Pick the initial state */
  current_state = pickAmongThree(chain->pi,rng);
  if (current_state == -1) {
    return -1;
  }
  trace->num_entries = 1;
  trace->dates = (long int*)calloc(1,sizeof(long int));
  trace->dates[0] = 0;
  trace->states = (char*)calloc(1,sizeof(char));
  trace->states[0] = current_state;
  /* Go through transitions */
  for (time_step = 1; time_step <= length; time_step++) {
    new_state = pickAmongThree(
              (double*)(&((chain->matrix)[current_state][0])), rng);
    if (new_state == -1) {
      return -1;
    }
    if (current_state != new_state) {
      trace->dates = (long int *)realloc(trace->dates, 
                            sizeof(long int)*(trace->num_entries+1));
      trace->dates[trace->num_entries] = time_step;
      trace->states = (char *)realloc(trace->states, 
                            sizeof(char)*(trace->num_entries+1));
      trace->states[trace->num_entries] = (hoststate_t)new_state;
      (trace->num_entries)++;
      current_state = new_state;
    }
  }

  /* Free random number generator */
  gsl_rng_free(rng);
  
  return 0;
}
/* buildTraceFromChainWeibull() */
static int buildTraceFromChainWeibull(int seed, long int length, double weibull_shape,
           markov_chain_t *chain, availability_trace_t *trace)
{
  gsl_rng *rng;
  int current_state;
  int time_step;
  double weibull_scale[3];


  /* Initialize random number generator */
  rng = gsl_rng_alloc(gsl_rng_taus2);
  gsl_rng_set(rng, seed);

  /* Initialize the trace */
  trace->num_entries = 0;
  trace->dates = NULL;
  trace->states = NULL;

  /* Compute the Weibull scale parameter for each state */
  int state;
  for (state=0; state<3; state++) {
    double mean_sojourn_time = 1.0 / (1.0 - chain->matrix[state][state]);
    weibull_scale[state] = mean_sojourn_time  / gsl_sf_gamma(1.0 + 1.0 / weibull_shape);
//    fprintf(stderr,"mean_sojourn_time: %lf pareto_shape: %lf   pareto_scale: %lf\n",
//		mean_sojourn_time,pareto_shape,pareto_scale[state]);
  }

  /* Pick the initial state */
  current_state = pickAmongThree(chain->pi,rng);
  if (current_state == -1) {
    return -1;
  }
  trace->num_entries = 1;
  trace->dates = (long int*)calloc(1,sizeof(long int));
  trace->dates[0] = 0;
  trace->states = (char*)calloc(1,sizeof(char));
  trace->dates[0] = current_state;

  /* Go through sojourns */
  time_step = 0;
  while (time_step <= length) {
    /* get an (integer) Weibull number */
    int weibull_number =  1+(int)(gsl_ran_weibull(rng, weibull_shape, weibull_scale[current_state]));
    //fprintf(stderr,"weibull number: %d\n",weibull_number);

    /* update the trace */
    trace->dates = (long int *)realloc(trace->dates, 
                          sizeof(long int)*(trace->num_entries+1));
    trace->dates[trace->num_entries] = time_step + weibull_number;

    trace->states = (char *)realloc(trace->states, 
                          sizeof(char)*(trace->num_entries+1));
    trace->states[trace->num_entries] = (hoststate_t)current_state;
    (trace->num_entries)++;
   
    /* increment the time step */
    time_step += weibull_number;

    /* pick a new state */
    current_state = pickAmongThree((double*)(&((chain->matrix)[current_state][0])), rng);
    if (current_state == -1) {
      return -1;
    }
  }

  /* Free random number generator */
  gsl_rng_free(rng);
  
  return 0;
}


/* buildTraceFromChainPareto() */
static int buildTraceFromChainPareto(int seed, long int length, double pareto_shape,
           markov_chain_t *chain, availability_trace_t *trace)
{
  gsl_rng *rng;
  int current_state;
  int time_step;
  double pareto_scale[3];


  /* Initialize random number generator */
  rng = gsl_rng_alloc(gsl_rng_taus2);
  gsl_rng_set(rng, seed);

  /* Initialize the trace */
  trace->num_entries = 0;
  trace->dates = NULL;
  trace->states = NULL;

  /* Compute the Pareto scale parameter for each state */
  int state;
  for (state=0; state<3; state++) {
    double mean_sojourn_time = 1.0 / (1.0 - chain->matrix[state][state]);
    pareto_scale[state] = mean_sojourn_time * (pareto_shape - 1.0) / pareto_shape;
  }

  /* Pick the initial state */
  current_state = pickAmongThree(chain->pi,rng);
  if (current_state == -1) {
    return -1;
  }
  trace->num_entries = 1;
  trace->dates = (long int*)calloc(1,sizeof(long int));
  trace->dates[0] = 0;
  trace->states = (char*)calloc(1,sizeof(char));
  trace->dates[0] = current_state;

  /* Go through sojourns */
  time_step = 0;
  while (time_step <= length) {
    /* get an (integer) Pareto number */
    int pareto_number =  1+(int)(gsl_ran_pareto(rng, pareto_shape, pareto_scale[current_state]));
    //fprintf(stderr,"pareto number: %d\n",pareto_number);

    /* update the trace */
    trace->dates = (long int *)realloc(trace->dates, 
                          sizeof(long int)*(trace->num_entries+1));
    trace->dates[trace->num_entries] = time_step + pareto_number;

    trace->states = (char *)realloc(trace->states, 
                          sizeof(char)*(trace->num_entries+1));
    trace->states[trace->num_entries] = (hoststate_t)current_state;
    (trace->num_entries)++;
   
    /* increment the time step */
    time_step += pareto_number;

    /* pick a new state */
    current_state = pickAmongThree((double*)(&((chain->matrix)[current_state][0])), rng);
    if (current_state == -1) {
      return -1;
    }
  }

  /* Free random number generator */
  gsl_rng_free(rng);
  
  return 0;
}


/* printTrace() */
void printTrace(availability_trace_t *trace)
{
  int i;
  char charstate[3]="URD";

  for (i=0; i < trace->num_entries; i++) {
    fprintf(stderr,"%ld %c\n",trace->dates[i],
                             charstate[(int)(trace->states[i])]); 
  }
  return;
}

/* buildChainFromTrace() */
int buildChainFromTrace(markov_chain_t *chain, availability_trace_t *trace)
{
  int occupancy_counts[3];
  int transition_counts[3][3];
  int i,j;

  /* Compute the "pi distribution" */
  for (i=0; i<3; i++)
    occupancy_counts[i] = 0;

  for (i=0; i<trace->num_entries-1; i++) {
    occupancy_counts[(int)(trace->states[i])] += 
                   trace->dates[i+1] - trace->dates[i];
  }
  for (i=0; i<3; i++)
    chain->pi[i] = ((double)occupancy_counts[i]) /
         ((double)(trace->dates[trace->num_entries-1]));
  
  /* Compute the "transition probability" */
  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      transition_counts[i][j] = 0;

  for (i=1; i<trace->num_entries; i++) {
    // Some time spent in state trace->states[i-1] 
    transition_counts[(int)(trace->states[i-1])][(int)(trace->states[i-1])] 
              += trace->dates[i] - trace->dates[i-1] -1;
    // One transition to state trace->states[i]
    transition_counts[(int)(trace->states[i-1])][(int)(trace->states[i])]++;
  }

  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      chain->matrix[i][j] = 
        ((double)(transition_counts[i][j]))/
                 ((double)occupancy_counts[i]);

//  fprintf(stderr,"UU_count=%ld/%ld\n",transition_counts[HOST_UP][HOST_UP],occupancy_counts[HOST_UP]);
//  fprintf(stderr,"PUU=%.10f\n",chain->matrix[HOST_UP][HOST_UP]);

  return 0;
}

/* printChain() */
void printChain(markov_chain_t *chain) 
{
  int i,j;

  for (i=0; i < 3; i++)  {
    fprintf(stderr,"  |");
    for (j=0; j < 3; j++) {
      fprintf(stderr,"%.2f ",chain->matrix[i][j]);
    }
    fprintf(stderr,"|\n");
  }
  fprintf(stderr,"  pi = [%.2f %.2f %.2f]\n",
                    chain->pi[0],chain->pi[1],chain->pi[2]);
}

/* parseTraceFile() */
int parseTraceFile(char *file, availability_trace_t *trace)
{
  long int first_date = -1;
  long int last_seen_date = -1;
  FILE *f;
  char line[256];
  int linecount=0;

  trace->num_entries = 0;
  trace->dates = NULL;
  trace->states = NULL;

  if (!(f = fopen(file,"r"))) {
    fprintf(stderr,"Cannot open trace file '%s'\n",file);
    return -1;
  }

  while (fgets(line,256,f)) {
    long int date;
    char charstate;
    char state=0;
    linecount++;
    if (sscanf(line,"%ld %c\n",&date,&charstate) != 2) {
      fprintf(stderr,"Error at line %d in trace file '%s'\n",
                     linecount,file);
      return -1;
    }
    switch (charstate) {
      case 'U':
        state = 0;
        break;
      case 'R':
        state = 1;
        break;
      case 'D':
        state = 2;
        break;
    }

    if (first_date == -1) {
      first_date = date;
    }

    if (((last_seen_date != -1) && (last_seen_date != date)) ||
        (last_seen_date == -1)) {
      trace->dates = (long int *)realloc(trace->dates, 
                       (trace->num_entries+1)*sizeof(long int));
      trace->states = (char *)realloc(trace->states, 
                       (trace->num_entries+1)*sizeof(char));
      trace->num_entries++;
    } 
    trace->dates[trace->num_entries-1] = date - first_date;
    trace->states[trace->num_entries-1] = state;
    last_seen_date = date;
  }
  fclose(f);
  return 0;
}
