#include <math.h>
#include "simulator.h"
#include <string.h>

static double *compute_pi_distribution();
static void reset_binary_counter(char *counter, int size);
static void print_binary_counter(char *counter, int size);
static void increment_binary_counter(char *counter, int size);
static double compute_transition_probability(char *counter1, char *counter2, int size);
static double compute_throughput_greedy(double *pi);
static double throughput_contribution(char *);

/* run_model() */
void run_model() 
{
  int i;
  double *pi;
  int dim = (int)pow(2.0,(double)global.num_hosts);
  double throughput_greedy=0.0;

  // compute the steady state distribution of the transition matrix
  // of dimension 2^numhosts
  pi = compute_pi_distribution(); 
  if (pi == NULL)
    return;

#if 0
  for (i=0; i<dim; i++)
    fprintf(stderr,"PI[%d] = %.3f\n",i,pi[i]); 
#endif

  // compute the greedy throughput

  throughput_greedy = compute_throughput_greedy(pi);

  // print output
  fprintf(stdout,"%.4f\n",throughput_greedy);
  
  free(pi);
  return;
}

/* compute_throughput_greedy() */
static double compute_throughput_greedy(double *pi)
{
  double throughput = 0.0;
  int dim = (int)pow(2.0,(double)global.num_hosts);
  char *counter;
  int i;

  if (global.migoverhead != 0.0) {
    fprintf(stderr,"Model with non-zero migration overhead not implemented!\n");
    return -1.0;
  }

  counter = (char *)calloc(global.num_hosts,sizeof(char));
  reset_binary_counter(counter,global.num_hosts);
  for (i=0; i < dim; i++) {
    throughput += pi[i] * throughput_contribution(counter);
    increment_binary_counter(counter,global.num_hosts); 
  }
  free(counter);
  return throughput;
}

/* throughput_contribution() */
static double throughput_contribution(char *counter)
{
  double contribution;
  int count;
  int i;
  // Right now, very simple: if the number of hosts
  // in the up state is larger than the number of
  // tasks, then contribution is host speed / task size
  count = 0;
  for (i=0; i<global.num_hosts; i++) {
    if (counter[i] == 1)
      count++;
  }
  if (count >= global.num_tasks)
    return (global.hosts[0].speed / global.tasks[0].cost);
  else
    return 0.0;
}

/* compute_pi_distribution() */
static double *compute_pi_distribution()
{
  int i,j;
  int dim = (int)pow(2.0,(double)global.num_hosts);
  double *trans_mat; 
  char *counter1;
  char *counter2;
  double *pi;

  // Allocate space for the transition matrix and the counters
  fprintf(stderr,"--> %d\n",dim*dim);
  trans_mat = (double *)calloc(dim*dim, sizeof(double));
  counter1 = (char *)calloc(global.num_hosts,sizeof(char));
  counter2 = (char *)calloc(global.num_hosts,sizeof(char));

  // TODO: THIS IS INSANE!!!
  { char *foo = (char *)calloc(10,sizeof(char)); }

  // Build matrix
  reset_binary_counter(counter1,global.num_hosts);
  for (i=0; i < dim; i++) {
    reset_binary_counter(counter2,global.num_hosts);
    for (j=0; j < dim; j++) {
      // Transpose the elements
      trans_mat[j*dim+i] = compute_transition_probability(counter1, counter2, global.num_hosts);
//      fprintf(stderr,"Probability of going from: '");
//      print_binary_counter(counter1,global.num_hosts);
//      fprintf(stderr,"' to '");
//      print_binary_counter(counter2,global.num_hosts);
//      fprintf(stderr,"' = %.3f\n", trans_mat[i][j]);
      increment_binary_counter(counter2,global.num_hosts); 
    }
    increment_binary_counter(counter1,global.num_hosts); 
  }
  free(counter1);
  free(counter2);

  

#if 0
  for (i=0; i < dim; i++) {
    for (j=0; j < dim; j++) {
      fprintf(stderr," %.2f",trans_mat[i*dim+j]);
    }
    fprintf(stderr,"\n");
  }
#endif

  // Call LAPACK 
  {
    char jobvl='V';
    char jobvr='N';
    double *wr = calloc(dim,sizeof(double));
    double *wi = calloc(dim,sizeof(double));
    double *vl = calloc(dim*dim,sizeof(double));
    int lwork = 6*dim;
    double *work = calloc(lwork,sizeof(double));
    int info=0;
    int i,j;
    double sum;

    dgeev_(&jobvl,&jobvr,&dim,trans_mat,&dim,wr,wi,vl,&dim,NULL,&dim,work,&lwork,&info);

    if (info) {
      fprintf(stderr,"Error: Cannot compute eigenvalues/eigenvectors of transition matrix (INFO=%d)\n",info);
      return NULL;
    }

    // find the index of eigenvalue 1
    for (i=0; i<dim; i++) {
      if ((fabs(wr[i]-1.0) < EPSILON) && (fabs(wi[i]) < EPSILON))
        break;
    }

    if (i == dim) {
      fprintf(stderr,"Error: Cannot find eigenvalue 1\n");
      return NULL;
    }
    
    // create the eigenvector
    pi = (double *)calloc(dim,sizeof(double*));
    sum = 0.0;
    for (j=0; j<dim; j++) {
      pi[j] = vl[i*dim+j];
      if (pi[j] < 0.0) 
        pi[j] = -pi[j];
      sum += pi[j];
    }

    // normalize the eigenvector
    for (j=0; j<dim; j++) {
      pi[j] /= sum;
    }
 
    // free stuff
    free(trans_mat);
    free(wr);
    free(wi);
    free(vl);
    free(work);
  }

  return pi;
}

/* reset binary counter */
static void reset_binary_counter(char *counter, int size)
{
  int i;
  for (i=0; i<size; i++)
    counter[i] = 0;
  return;
}

/* increment_binary_counter */
static void increment_binary_counter(char *counter, int size)
{
  int i;

  for (i=size-1; i>= 0; i--) {
    counter[i] = 1 - counter[i];
    if (counter[i] == 1)
      break;
  }
  return;
}

/* print_binary_counter */
static void print_binary_counter(char *counter, int size)
{
  int i;
  for (i=0; i<size; i++) 
    fprintf(stderr,"%d",counter[i]);
  return;
}

static double compute_transition_probability(char *counter1, char *counter2, int size)
{
  int i;
  double prob = 1.0;
  

  for (i=0; i<size; i++) {
    switch(counter1[i]) {
      case 0:
        switch(counter2[i]) {
          case 0:
            prob *= global.hosts[i].p[0];
            break;
          case 1: 
            prob *= 1-global.hosts[i].p[0];
            break;
        }
        break;

      case 1:
        switch(counter2[i]) {
          case 0:
            prob *= 1-global.hosts[i].p[1];
            break;
          case 1: 
            prob *= global.hosts[i].p[1];
            break;
        }
        break;
    }   
  }
  return prob;
}
