#include <string.h>
#include <math.h>
#include <gsl/gsl_integration.h>
#include "simulator.h"

#define NOT_SCALED 0
#define SCALED 1
#define NOT_EXPECTED 0
#define EXPECTED 1
#define NOT_CONDITIONAL 0
#define CONDITIONAL 1
#define NOT_NAIVE 0
#define NAIVE 1

/* Processor scheduling functions */
static host_t *pick_random_host_up(const char *weight, host_t **);
static host_t *pick_MCT_host_up(char, char, host_t **);
static host_t *pick_UD_host_up(char, char, host_t **);
static host_t *pick_LW_host_up(char, char, host_t **);
static host_t *pick_RATE_host_up(char, char, char, char, host_t **);
static host_t *pick_inc_host_up(const char *, host_t **, task_t *);

/* Utility functions */
static long compute_MCT_indep_task_ct(host_t *host, char scaled, int num_new_tasks);
static long compute_current_comp_ct(host_t *host);
static long compute_current_comm_ct(host_t *host);
static char will_host_communicate_in_the_future(host_t *host);
static double compute_contention_factor(void);
static double compute_expected_ct(host_t *, long);
static double compute_expected_rate(host_t *, double, char, char, char);
static double compute_erate(double, double, double, double, double, double, int, int);
static long communication_time_coupled(host_t *host);
static double weight_on_host(host_t *host);
static double eigenval1(host_t *host);
static double eigenval2(host_t *host);
static double alpha1(host_t *host);
static double alpha2(host_t *host);
static double compute_expected_uS_coupled(host_t **set_hosts, int set_host_size,double precision);
static double compute_expected_AS_coupled(host_t **set_hosts, int set_host_size,double precision);
static double compute_expected_Pfanny_coupled(host_t **set_hosts, int set_host_size, double precision, int weight);
static double compute_expected_Efanny_coupled(host_t **set_hosts, int set_host_size, double precision, int weight);
static void create_clones(void);
static int how_many_allowed_clones(void);
static int pick_task_to_clone(void);
static void create_task_clone(int task_index);
static int get_num_of_clones(int task_index);
static host_t **get_hosts_with_clone(task_t *task);
static int is_host_in_list(host_t *host, host_t **dont_pick);
static double compute_proba_success(int task_index);
static double coupled_success_proba(void);
static double coupled_exp_completion_time(void);
static double P_nD(host_t*,long);
static double function_rate_integral_1(double x, void *p);
static double function_rate_integral_2(double x, void *p);
static double function_rate_integral_3(double x, void *p);

/* Helper structures */
struct params_function_rate_integral {
  double a;
  double b;
  double c;
};

typedef enum {
  RANDOM=0,
  WEIGHTED_RANDOM_Puu,
  WEIGHTED_RANDOM_Puuw,
  WEIGHTED_RANDOM_Pfanny,
  WEIGHTED_RANDOM_Pfannyw,
  WEIGHTED_RANDOM_PIu,
  WEIGHTED_RANDOM_PIuw,
  WEIGHTED_RANDOM_PIuf,
  WEIGHTED_RANDOM_PIufw,
  MCT_v1,
  MCT_v2,
  LW_v1,
  LW_v2,
  UD_v1,
  UD_v2,
  RATE_v1,
  RATE_v2,
  EMCT_v1,
  EMCT_v2,
  ELW_v1,
  ELW_v2,
  EUD_v1,
  EUD_v2,
  ERATE_v1,
  ERATE_v2,
  ENCRATE_v1,
  ENCRATE_v2,
  ENCRATENAIVE_v1,
  ENCRATENAIVE_v2,
  IE,
  IP,
  IR,
  IY,
  RANDOM_PP,
  WEIGHTED_RANDOM_Puu_PP,
  WEIGHTED_RANDOM_Puuw_PP,
  WEIGHTED_RANDOM_Pfanny_PP,
  WEIGHTED_RANDOM_Pfannyw_PP,
  WEIGHTED_RANDOM_PIu_PP,
  WEIGHTED_RANDOM_PIuw_PP,
  WEIGHTED_RANDOM_PIuf_PP,
  WEIGHTED_RANDOM_PIufw_PP,
  MCT_v1_PP,
  MCT_v2_PP,
  LW_v1_PP,
  LW_v2_PP,
  UD_v1_PP,
  UD_v2_PP,
  EMCT_v1_PP,
  EMCT_v2_PP,
  ELW_v1_PP,
  ELW_v2_PP,
  EUD_v1_PP,
  EUD_v2_PP,
  IE_PP,
  IP_PP,
  IR_PP,
  IY_PP,
  RANDOM_PE,
  WEIGHTED_RANDOM_Puu_PE,
  WEIGHTED_RANDOM_Puuw_PE,
  WEIGHTED_RANDOM_Pfanny_PE,
  WEIGHTED_RANDOM_Pfannyw_PE,
  WEIGHTED_RANDOM_PIu_PE,
  WEIGHTED_RANDOM_PIuw_PE,
  WEIGHTED_RANDOM_PIuf_PE,
  WEIGHTED_RANDOM_PIufw_PE,
  MCT_v1_PE,
  MCT_v2_PE,
  LW_v1_PE,
  LW_v2_PE,
  UD_v1_PE,
  UD_v2_PE,
  EMCT_v1_PE,
  EMCT_v2_PE,
  ELW_v1_PE,
  ELW_v2_PE,
  EUD_v1_PE,
  EUD_v2_PE,
  IE_PE,
  IP_PE,
  IR_PE,
  IY_PE,
  RANDOM_PR,
  WEIGHTED_RANDOM_Puu_PR,
  WEIGHTED_RANDOM_Puuw_PR,
  WEIGHTED_RANDOM_Pfanny_PR,
  WEIGHTED_RANDOM_Pfannyw_PR,
  WEIGHTED_RANDOM_PIu_PR,
  WEIGHTED_RANDOM_PIuw_PR,
  WEIGHTED_RANDOM_PIuf_PR,
  WEIGHTED_RANDOM_PIufw_PR,
  MCT_v1_PR,
  MCT_v2_PR,
  LW_v1_PR,
  LW_v2_PR,
  UD_v1_PR,
  UD_v2_PR,
  EMCT_v1_PR,
  EMCT_v2_PR,
  ELW_v1_PR,
  ELW_v2_PR,
  EUD_v1_PR,
  EUD_v2_PR,
  IE_PR,
  IP_PR,
  IR_PR,
  IY_PR,
  RANDOM_PY,
  WEIGHTED_RANDOM_Puu_PY,
  WEIGHTED_RANDOM_Puuw_PY,
  WEIGHTED_RANDOM_Pfanny_PY,
  WEIGHTED_RANDOM_Pfannyw_PY,
  WEIGHTED_RANDOM_PIu_PY,
  WEIGHTED_RANDOM_PIuw_PY,
  WEIGHTED_RANDOM_PIuf_PY,
  WEIGHTED_RANDOM_PIufw_PY,
  MCT_v1_PY,
  MCT_v2_PY,
  LW_v1_PY,
  LW_v2_PY,
  UD_v1_PY,
  UD_v2_PY,
  EMCT_v1_PY,
  EMCT_v2_PY,
  ELW_v1_PY,
  ELW_v2_PY,
  EUD_v1_PY,
  EUD_v2_PY,
  IE_PY,
  IP_PY,
  IR_PY,
  IY_PY
} host_selection_type_t;

static int schedule_indep(config_t *config,
                            host_selection_type_t host_selection_type);

static int schedule_coupled_proactif(config_t *config,
                            host_selection_type_t host_selection_type, const char *criterion);

static int schedule_coupled(config_t *config,
                            host_selection_type_t host_selection_type);

/* schedule_random() */
int schedule_indep_random(config_t *config) {
  return schedule_indep(config, RANDOM);
}

/* schedule_indep_weighted_random_Puu() */
int schedule_indep_weighted_random_Puu(config_t *config) {
  return schedule_indep(config, WEIGHTED_RANDOM_Puu);
}

/* schedule_indep_weighted_random_Puuw() */
int schedule_indep_weighted_random_Puuw(config_t *config) {
  return schedule_indep(config, WEIGHTED_RANDOM_Puuw);
}

/* schedule_indep_weighted_random_Pfanny() */
int schedule_indep_weighted_random_Pfanny(config_t *config) {
  return schedule_indep(config, WEIGHTED_RANDOM_Pfanny);
}

/* schedule_indep_weighted_random_Pfannyw() */
int schedule_indep_weighted_random_Pfannyw(config_t *config) {
  return schedule_indep(config, WEIGHTED_RANDOM_Pfannyw);
}

/* schedule_indep_weighted_random_PIu() */
int schedule_indep_weighted_random_PIu(config_t *config) {
  return schedule_indep(config, WEIGHTED_RANDOM_PIu);
}

/* schedule_indep_weighted_random_PIuw() */
int schedule_indep_weighted_random_PIuw(config_t *config) {
  return schedule_indep(config, WEIGHTED_RANDOM_PIuw);
}

/* schedule_indep_weighted_random_PIuf() */
int schedule_indep_weighted_random_PIuf(config_t *config) {
  return schedule_indep(config, WEIGHTED_RANDOM_PIuf);
}

/* schedule_indep_weighted_random_PIufw() */
int schedule_indep_weighted_random_PIufw(config_t *config) {
  return schedule_indep(config, WEIGHTED_RANDOM_PIufw);
}

/* schedule_MCT_indep_v1() */
int schedule_indep_MCT_v1(config_t *config) {
  return schedule_indep(config, MCT_v1);
}

/* schedule_MCT_indep_v2() */
int schedule_indep_MCT_v2(config_t *config) {
  return schedule_indep(config, MCT_v2);
}

/* schedule_UD_indep_v1() */
int schedule_indep_UD_v1(config_t *config) {
  return schedule_indep(config, UD_v1);
}

/* schedule_UD_indep_v2() */
int schedule_indep_UD_v2(config_t *config) {
  return schedule_indep(config, UD_v2);
}

/* schedule_LW_indep_v1() */
int schedule_indep_LW_v1(config_t *config) {
  return schedule_indep(config, LW_v1);
}

/* schedule_LW_indep_v2() */
int schedule_indep_LW_v2(config_t *config) {
  return schedule_indep(config, LW_v2);
}

/* schedule_RATE_indep_v1() */
int schedule_indep_RATE_v1(config_t *config) {
  return schedule_indep(config, RATE_v1);
}

/* schedule_RATE_indep_v2() */
int schedule_indep_RATE_v2(config_t *config) {
  return schedule_indep(config, RATE_v2);
}

/* schedule_EMCT_indep_v1() */
int schedule_indep_EMCT_v1(config_t *config) {
  return schedule_indep(config, EMCT_v1);
}

/* schedule_EMCT_indep_v2() */
int schedule_indep_EMCT_v2(config_t *config) {
  return schedule_indep(config, EMCT_v2);
}

/* schedule_EUD_indep_v1() */
int schedule_indep_EUD_v1(config_t *config) {
  return schedule_indep(config, EUD_v1);
}

/* schedule_EUD_indep_v2() */
int schedule_indep_EUD_v2(config_t *config) {
  return schedule_indep(config, EUD_v2);
}

/* schedule_ELW_indep_v1() */
int schedule_indep_ELW_v1(config_t *config) {
  return schedule_indep(config, ELW_v1);
}

/* schedule_ELW_indep_v2() */
int schedule_indep_ELW_v2(config_t *config) {
  return schedule_indep(config, ELW_v2);
}

/* schedule_ERATE_indep_v1() */
int schedule_indep_ERATE_v1(config_t *config) {
  return schedule_indep(config, ERATE_v1);
}

/* schedule_ERATE_indep_v2() */
int schedule_indep_ERATE_v2(config_t *config) {
  return schedule_indep(config, ERATE_v2);
}

/* schedule_ENCRATE_indep_v1() */
int schedule_indep_ENCRATE_v1(config_t *config) {
  return schedule_indep(config, ENCRATENAIVE_v1);
}

/* schedule_ENCRATE_indep_v2() */
int schedule_indep_ENCRATE_v2(config_t *config) {
  return schedule_indep(config, ENCRATENAIVE_v2);
}

/* schedule_ENCRATENAIVE_indep_v1() */
int schedule_indep_ENCRATENAIVE_v1(config_t *config) {
  return schedule_indep(config, ENCRATENAIVE_v1);
}

/* schedule_ENCRATENAIVE_indep_v2() */
int schedule_indep_ENCRATENAIVE_v2(config_t *config) {
  return schedule_indep(config, ENCRATENAIVE_v2);
}

/* Main schedule_indep() function */
static int schedule_indep(config_t *config,
                            host_selection_type_t host_selection_type) 
{
  int i,j;

  /* Reset all the host counters that will be used to track which procs
     were "picked" first for tasks, so as to prioritize communications
     accordingly (proc picked first -> new communication to it has priority */
  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = global.platform.hosts[i];
    host->picking_order = global.config->num_tasks; // worst ranking + 1
  }

  /* Create clones */
  create_clones();
  
  /* Assign all tasks that are unassigned, according to some criterion */
  /* Consider all original tasks first, then all 1st clones, etc. */
  for (j=0; j < global.platform.num_hosts; j++) {
    for (i=0; i < global.application.num_tasks; i++) {
      task_t *task =config->tasks[i * global.platform.num_hosts + j];
      host_t *host = NULL;

      /* If the task is not activated, forget it */
      if (!task->activated)
        continue;
  
      /* If the task is assigned or done, leave it where it is */
      if ((task->assigned_host) || (task->done))
        continue;
  
      /* Build the list of hosts to avoid because a clone of
         the task's already scheduled/running on that host */
      host_t **dont_pick = get_hosts_with_clone(task);
      /* pick a host */
      switch(host_selection_type) {
  
        case RANDOM: 
          host = pick_random_host_up("none",dont_pick);
          break;
  
        case WEIGHTED_RANDOM_Puu:
          host = pick_random_host_up("Puu",dont_pick);
          break;
  
        case WEIGHTED_RANDOM_Puuw:
          host = pick_random_host_up("Puuw",dont_pick);
          break;
  
        case WEIGHTED_RANDOM_Pfanny:
          host = pick_random_host_up("Pfanny",dont_pick);
          break;
  
        case WEIGHTED_RANDOM_Pfannyw:
          host = pick_random_host_up("Pfannyw",dont_pick);
          break;
  
        case WEIGHTED_RANDOM_PIu:
          host = pick_random_host_up("PIu",dont_pick);
          break;
  
        case WEIGHTED_RANDOM_PIuw:
          host = pick_random_host_up("PIuw",dont_pick);
          break;
  
        case WEIGHTED_RANDOM_PIuf:
          host = pick_random_host_up("PIuf",dont_pick);
          break;
  
        case WEIGHTED_RANDOM_PIufw:
          host = pick_random_host_up("PIufw",dont_pick);
          break;
  
        case MCT_v1:
	  host = pick_MCT_host_up(NOT_SCALED,NOT_EXPECTED,dont_pick);
	  break;
	  
        case MCT_v2:
	  host = pick_MCT_host_up(SCALED,NOT_EXPECTED,dont_pick);
	  break;

        case LW_v1:
	  host = pick_LW_host_up(NOT_SCALED,NOT_EXPECTED,dont_pick);
	  break;
	  
        case LW_v2:
	  host = pick_LW_host_up(SCALED,NOT_EXPECTED,dont_pick);
	  break;
  
        case UD_v1:
	  host = pick_UD_host_up(NOT_SCALED,NOT_EXPECTED,dont_pick);
	  break;
	  
        case UD_v2:
	  host = pick_UD_host_up(SCALED,NOT_EXPECTED,dont_pick);
	  break;

        case RATE_v1:
	  host = pick_RATE_host_up(NOT_SCALED,NOT_EXPECTED,
                                   CONDITIONAL,NOT_NAIVE,dont_pick);
	  break;
	  
        case RATE_v2:
	  host = pick_RATE_host_up(SCALED,NOT_EXPECTED,
                                   CONDITIONAL,NOT_NAIVE,dont_pick);
	       
        case EMCT_v1:
	  host = pick_MCT_host_up(NOT_SCALED,EXPECTED,dont_pick);
	  break;
	  
        case EMCT_v2:
	  host = pick_MCT_host_up(SCALED,EXPECTED,dont_pick);
	  break;

        case ELW_v1:
	  host = pick_LW_host_up(NOT_SCALED,EXPECTED,dont_pick);
	  break;
	  
        case ELW_v2:
	  host = pick_LW_host_up(SCALED,EXPECTED,dont_pick);
	  break;
  
        case EUD_v1:
	  host = pick_UD_host_up(NOT_SCALED,EXPECTED,dont_pick);
	  break;
	  
        case EUD_v2:
	  host = pick_UD_host_up(SCALED,EXPECTED,dont_pick);
	  break;

        case ERATE_v1:
	  host = pick_RATE_host_up(NOT_SCALED,EXPECTED,
                                   CONDITIONAL,NOT_NAIVE,dont_pick);
	  break;
	  
        case ERATE_v2:
	  host = pick_RATE_host_up(SCALED,EXPECTED,
                                   CONDITIONAL,NOT_NAIVE,dont_pick);
	  break;

        case ENCRATE_v1:
	  host = pick_RATE_host_up(NOT_SCALED,EXPECTED,
                                   NOT_CONDITIONAL,NOT_NAIVE,dont_pick);
	  break;
	  
        case ENCRATE_v2:
	  host = pick_RATE_host_up(SCALED,EXPECTED,
                                   NOT_CONDITIONAL,NOT_NAIVE,dont_pick);
	  break;

        case ENCRATENAIVE_v1:
	  host = pick_RATE_host_up(NOT_SCALED,EXPECTED,
                                   NOT_CONDITIONAL,NAIVE,dont_pick);
	  break;
	  
        case ENCRATENAIVE_v2:
	  host = pick_RATE_host_up(SCALED,EXPECTED,
                                   NOT_CONDITIONAL,NAIVE,dont_pick);
	  break;

	default:
	  fprintf(stderr,"heuristic coupled in model indep\n");
	  exit(1);
      }
    
      free(dont_pick);

      /* Assign the task to host if a host was found */
      if (host) {
        if (host->picking_order > i) {
           host->picking_order = i;
        }
        assign_task_to_host(task,host);
      }
    }
  }
  return 0;
}

/* pick_random_host_up()
 * "none" (WEIGHTED_RANDOM): weight = 1
 * "Puu" (WEIGHTED_RANDOM_Puu): weight = P_{u,u}
 * "Puuw" (WEIGHTED_RANDOM_Puuw): weight = P_{u,u}/w
 * "Pfanny" (WEIGHTED_RANDOM_Pfanny): weight = P_{fanny}
 * "Pfannyw" (WEIGHTED_RANDOM_Pfannyw): weight = P_{fanny}/w
 * "PIu" (WEIGHTED_RANDOM_PIu): weight = \pi_{u}
 * "PIuw" (WEIGHTED_RANDOM_PIuw): weight = \pi_{u}/w
 * "PIfanny" (WEIGHTED_RANDOM_PIuf): weight = (\pi_{u} + \pi_{f})/w
 * "PIfannyw" (WEIGHTED_RANDOM_PIufw): weight = (\pi_{u} + \pi_{f})/w
 */
static host_t *pick_random_host_up(const char *weight, host_t **dont_pick) {
  int *up_host_indices = NULL;
  int found_index;
  int num_up_hosts = 0;
  int up_host_index;
  int j;

  /* Record the indices of all up hosts */
  for (j=0; j < global.platform.num_hosts; j++) {
    host_t *host = global.platform.hosts[j];

    /* If host is in the "don't pick" list, forget it */
    if (is_host_in_list(host, dont_pick)) { 
      continue;
    }

    if (STATE(host) == HOST_UP) {
      up_host_indices = (int*)realloc(up_host_indices, (num_up_hosts+1)*sizeof(int));
      up_host_indices[num_up_hosts] = j;
      num_up_hosts++;
    }
  }
 
  if (num_up_hosts == 0)
    return NULL;

    /* WEIGHTED_RANDOM */
  if (!strcmp(weight,"none")) {
    double *weights;
    int i;

    weights = (double *)calloc(num_up_hosts,sizeof(double));
    for (i=0; i < num_up_hosts; i++)  {
      weights[i] = 1.0;
    }
    up_host_index = pick_weighted(weights, num_up_hosts);
    free(weights);

    /* WEIGHTED_Puu */
  } else if (!strcmp(weight,"Puu")) {
    double *weights;
    int i;

    weights = (double *)calloc(num_up_hosts,sizeof(double));
    for (i=0; i < num_up_hosts; i++)  {
      weights[i] = global.platform.hosts[up_host_indices[i]]->chain.matrix[HOST_UP][HOST_UP];
    }
    up_host_index = pick_weighted(weights, num_up_hosts);
    free(weights);

    /* WEIGHTED_Puuw */
  } else if (!strcmp(weight,"Puuw")) {
    double *weights;
    int i;

    weights = (double *)calloc(num_up_hosts,sizeof(double));
    for (i=0; i < num_up_hosts; i++)  {
      weights[i] = global.platform.hosts[up_host_indices[i]]->chain.matrix[HOST_UP][HOST_UP] /
                   (double) (global.platform.hosts[up_host_indices[i]]->w);
    }
    up_host_index = pick_weighted(weights, num_up_hosts);
    free(weights);

  /* WEIGHTED_Pfanny */
  } else if (!strcmp(weight,"Pfanny")) {
    double *weights;
    int i;

    weights = (double *)calloc(num_up_hosts,sizeof(double));
    for (i=0; i < num_up_hosts; i++)  {
      weights[i] = global.platform.hosts[up_host_indices[i]]->Pfanny;
    }
    up_host_index = pick_weighted(weights, num_up_hosts);
    free(weights);

    /* WEIGHTED_Pfannyw */
  } else if (!strcmp(weight,"Pfannyw")) {
    double *weights;
    int i;

    weights = (double *)calloc(num_up_hosts,sizeof(double));
    for (i=0; i < num_up_hosts; i++)  {
      weights[i] = global.platform.hosts[up_host_indices[i]]->Pfanny /
                   (double) (global.platform.hosts[up_host_indices[i]]->w);
    }
    up_host_index = pick_weighted(weights, num_up_hosts);
    free(weights);

  /* WEIGHTED_PIu */
  } else if (!strcmp(weight,"PIu")) {
    double *weights;
    int i;

    weights = (double *)calloc(num_up_hosts,sizeof(double));
    for (i=0; i < num_up_hosts; i++)  {
      weights[i] = global.platform.hosts[up_host_indices[i]]->chain.pi[HOST_UP];
    }
    up_host_index = pick_weighted(weights, num_up_hosts);
    free(weights);

  /* WEIGHTED_PIuw */
  } else if (!strcmp(weight,"PIuw")) {
    double *weights;
    int i;

    weights = (double *)calloc(num_up_hosts,sizeof(double));
    for (i=0; i < num_up_hosts; i++)  {
      weights[i] = (1.0 - global.platform.hosts[up_host_indices[i]]->chain.pi[HOST_UP]) /
                   (double) (global.platform.hosts[up_host_indices[i]]->w);
    }
    up_host_index = pick_weighted(weights, num_up_hosts);
    free(weights);

  /* WEIGHTED_PIuf */
  } else if (!strcmp(weight,"PIuf")) {
    double *weights;
    int i;

    weights = (double *)calloc(num_up_hosts,sizeof(double));
    for (i=0; i < num_up_hosts; i++)  {
      weights[i] = 1.0 - global.platform.hosts[up_host_indices[i]]->chain.pi[HOST_DOWN];
    }
    up_host_index = pick_weighted(weights, num_up_hosts);
    free(weights);

  /* WEIGHTED_PIuw */
  } else if (!strcmp(weight,"PIufw")) {
    double *weights;
    int i;

    weights = (double *)calloc(num_up_hosts,sizeof(double));
    for (i=0; i < num_up_hosts; i++)  {
      weights[i] = (1.0 - global.platform.hosts[up_host_indices[i]]->chain.pi[HOST_DOWN]) /
                   (double) (global.platform.hosts[up_host_indices[i]]->w);
    }
    up_host_index = pick_weighted(weights, num_up_hosts);
    free(weights);

  } else {
    fprintf(stderr,"pick_random_host_up(): invalid weight specification\n");
    exit(1);
  }

  found_index = up_host_indices[up_host_index];
  free(up_host_indices);

  return global.platform.hosts[found_index];
}

/* pick_MCT_host_up() */
static host_t *pick_MCT_host_up(char scaled, char expected, host_t **dont_pick)
{
  int j;

  host_t *picked_host = NULL;
  long picked_host_ct; 

  for (j=0; j < global.platform.num_hosts; j++) {
    host_t *host = global.platform.hosts[j];
    long host_ct;

    /* If host is in the "don't pick" list, forget it */
    if (is_host_in_list(host, dont_pick)) { 
      continue;
    }

    /* If host isn't up, move on */
    if (STATE(host) != HOST_UP)
      continue;

    /* Compute the CT account for all non-started-tasks and a new one */
    host_ct = compute_MCT_indep_task_ct(host, scaled, 
                         num_tasks_not_yet_started_on_host(host) + 1);

    if (expected) { // If Expected, then apply the Pfanny transformation
      host_ct = compute_expected_ct(host, host_ct);
    }

    /* Is it a better host? */
    if ((picked_host == NULL) || (host_ct < picked_host_ct)) {
      picked_host = host;
      picked_host_ct = host_ct;
    } else if (host_ct == picked_host_ct) {
      /* Breaking a tie */
      if (host->chain.matrix[HOST_UP][HOST_UP] > 
        picked_host->chain.matrix[HOST_UP][HOST_UP]) {
        picked_host = host;
        picked_host_ct = host_ct;
      }
    }
  }
//  fprintf(stderr,"\tPicked host#%d\n",picked_host->id);
  return picked_host;
}

/* pick_UD_host_up() */
static host_t *pick_UD_host_up(char scaled, char expected, host_t **dont_pick)
{
  int j;

  host_t *picked_host = NULL;
  double picked_host_ud;
  double host_ud; 

  for (j=0; j < global.platform.num_hosts; j++) {
    host_t *host = global.platform.hosts[j];

    /* If host is in the "don't pick" list, forget it */
    if (is_host_in_list(host, dont_pick)) { 
      continue;
    }

    /* If host isn't up, move on */
    if (STATE(host) != HOST_UP)
      continue;

    double probability = host->PUD_helper;

    /* Compute the CT accounting for all non-started tasks + 1 */
    long host_ct = compute_MCT_indep_task_ct(host, scaled, 
                          num_tasks_not_yet_started_on_host(host) + 1);
    if (!expected) {
      host_ud = pow(probability, host_ct - global.current_date - 2);
    } else {
      host_ud = pow( probability, (compute_expected_ct(host, host_ct) - global.current_date)-2);
    }

    probability *= (1.0 - host->chain.matrix[HOST_UP][HOST_DOWN]);

    /* Is it a better host? */
    if ((picked_host == NULL) || (host_ud > picked_host_ud)) {
      picked_host = host;
      picked_host_ud = host_ud;
    } else if (host_ud == picked_host_ud) {
      /* Breaking a tie */
      if (host->chain.matrix[HOST_UP][HOST_UP] > 
        picked_host->chain.matrix[HOST_UP][HOST_UP]) {
        picked_host = host;
        picked_host_ud = host_ud;
      }
    }
  }
  return picked_host;
}

/* pick_LW_host_up() */
static host_t *pick_LW_host_up(char scaled, char expected, host_t **dont_pick)
{
  int j;

  host_t *picked_host = NULL;
  double picked_host_lw;
  double host_lw; 

  for (j=0; j < global.platform.num_hosts; j++) {
    host_t *host = global.platform.hosts[j];

    /* If host is in the "don't pick" list, forget it */
    if (is_host_in_list(host, dont_pick)) { 
      continue;
    }

    /* If host isn't up, move on */
    if (STATE(host) != HOST_UP)
      continue;

    double probability = host->PLW_helper;
                  
    /* Compute the CT accounting for non-started tasks + 1 */
    long host_ct = compute_MCT_indep_task_ct(host, scaled, 
                           num_tasks_not_yet_started_on_host(host) + 1); 
    if (!expected) {
      host_lw = pow(probability, host_ct - global.current_date);
    } else {
      host_lw = pow(probability, 
                   (compute_expected_ct(host, host_ct) - global.current_date)-1);
    }

    /* Is it a better host? */
    if ((picked_host == NULL) || (host_lw > picked_host_lw)) {
      picked_host = host;
      picked_host_lw = host_lw;
    } else if (host_lw == picked_host_lw) {
      /* Breaking a tie */
      if (host->chain.matrix[HOST_UP][HOST_UP] > 
        picked_host->chain.matrix[HOST_UP][HOST_UP]) {
        picked_host = host;
        picked_host_lw = host_lw;
      }
    }
  }
  return picked_host;
}

/* pick_RATE_host_up() */
static host_t *pick_RATE_host_up(char scaled, char expected, 
                                 char conditional, char naive, 
                                 host_t **dont_pick)
{
  int j;

  double *host_rates = (double *)calloc(global.platform.num_hosts,sizeof(double));
  for (j=0; j < global.platform.num_hosts; j++) {
    host_rates[j] = -1.0;
  }

  double current_platform_rate = 0;

  for (j=0; j < global.platform.num_hosts; j++) {
    host_t *host = global.platform.hosts[j];

    /* If host is in the "don't pick" list, move on */
    if (is_host_in_list(host, dont_pick)) 
      continue;

    /* If host isn't up, move on */
    if (STATE(host) != HOST_UP)
      continue;

    /* Compute the CT accounting for non-started tasks + 1 */
    long host_ct = compute_MCT_indep_task_ct(host, scaled, 
                         num_tasks_not_yet_started_on_host(host) + 1); 

    if (expected == EXPECTED) {
      host_rates[j] = compute_expected_rate(host, 
                   (double)(host_ct - global.current_date), conditional, naive, scaled);
    } else {
      host_rates[j] = (1+num_tasks_not_yet_completed_on_host(host)) / 
                           (host_ct - global.current_date);
    }

    current_platform_rate += host_rates[j];
  }


  /* Find the host that, if assigned the current extra task, will maximize the platform's rate */
  double max_platform_rate = -1.0;
  
  host_t *picked_host = NULL;
  for (j=0; j < global.platform.num_hosts; j++) {
    host_t *host = global.platform.hosts[j];

    /* If the host was not a contender, mov on */
    if (host_rates[j] < 0) 
      continue;

    if (picked_host == NULL)  { /* first host */
      max_platform_rate = current_platform_rate + host_rates[j];
      picked_host = host;
    } else if (current_platform_rate + host_rates[j] > max_platform_rate) { /* better host */
      max_platform_rate = current_platform_rate + host_rates[j];
      picked_host = host;
    } else if (current_platform_rate + host_rates[j] == max_platform_rate) { /* tie */
      if (host->chain.matrix[HOST_UP][HOST_UP] >
        picked_host->chain.matrix[HOST_UP][HOST_UP]) {
        max_platform_rate = current_platform_rate + host_rates[j];
        picked_host = host;
      }
    }
  }

//  if (picked_host)
//    fprintf(stderr,"** Picked host#%d (platform_rate=%lf) **\n",picked_host->id,max_platform_rate);

  return picked_host;
}



/* pick_weighted(double *weights, int n) 
 * if weights = NULL, then uniform
 */
int pick_weighted(double *weights, int n)
{
  int i;
  double sum = 0.0;
  double draw;

  if (weights) {
    for (i=0; i < n; i++)   {
      sum += weights[i];
    }
  } else {
    sum = (double)n;
  }

  draw = ((double)rand() / (double)RAND_MAX) * sum;

  sum = 0.0;
  for (i=0; i < n; i++) {
    if (weights) {
      sum += weights[i];
    } else {
      sum += 1.0;
    }
    if (sum > draw)
      return i;
  }
  
  return n-1; 
}

/* compute_MCT_indep_task_ct() 
 *
 *   Computes the completion time of a host as defined
 *   by the MCT_indep heuristic (see paper), assuming there is a new
 *   task put on the new host
 *     scaled = NOT_SCALED: No accounting for contention
 *     scaled = SCALED: Accounting for contention with a correcting factor
 * 
 *   This doesn't do anything clever about understanding overlaps
 */
static long compute_MCT_indep_task_ct(host_t *host, char scaled, int num_new_tasks)
{
  long scaled_Tdata;
  long mct;
  long comm_ct, comp_ct;
 
  if (scaled == NOT_SCALED) {
    scaled_Tdata = global.application.Tdata;
  } else {
    scaled_Tdata = (long)(ceil((double)global.application.Tdata * 
                                compute_contention_factor()));
  }

  comm_ct = compute_current_comm_ct(host);
  comp_ct = compute_current_comp_ct(host);

  // Compute the MCT iteratively
  int i;
  int num_tasks_to_do = num_new_tasks;  // Note the +1
  for (i=0; i < num_tasks_to_do; i++) {
    comm_ct += scaled_Tdata;
    if(global.application.mode == MODE_COUPLED)
    {comp_ct +=host->w;}
    else{
    comp_ct = MAX(comm_ct + host->w, comp_ct + host->w);}
  }

  /* Coarse approximation - NOT USED ANYMORE */
/*
  mct = MAX(comm_ct,comp_ct) +  scaled_Tdata + 
        num_tasks_not_yet_started_on_host(host) * MAX(scaled_Tdata, host->w) + 
        host->w;
*/
  if(global.application.mode == MODE_COUPLED)
  {mct=comm_ct+comp_ct;}
  else{
  mct = MAX(comm_ct, comp_ct);}
  return mct;

//  fprintf(stderr,"mct=%ld\n",mct);
//  fprintf(stderr,"\tscaled_Tdata=%ld\n",scaled_Tdata);
//  fprintf(stderr,"\tnum_tasks_not_yet_started=%d\n",num_tasks_not_yet_started_on_host(host));
//  fprintf(stderr,"\thost->w=%d\n",host->w);
}

/* compute_current_comm_ct() 
 *  Computes the time at which currently 
 *  RUNNING (not scheduled) communication finishes.
 *
 */
static long compute_current_comm_ct(host_t *host)
{
  int i;
  int sanity_check;
  long current_comm_ct;

  /* Compute the time at which the (single) currently running 
     communication will complete */ 
  current_comm_ct = global.current_date + host->time_to_prog_download;
  sanity_check = 0;
  for (i=0; i < global.config->num_tasks; i++) {
    task_t *task = global.config->tasks[i];
    /* If the task isn't activated, forget it */
    if (!task->activated)
      continue;
    /* If task isn't scheduled on host, forget it */
    if (task->assigned_host != host)
      continue;
    if ((task->time_to_data_download < global.application.Tdata) && 
        (task->time_to_data_download > 0)) {
      current_comm_ct += task->time_to_data_download;
      sanity_check++;
     /* Sanity check: we shouldn't do this more than once!! */
      if (sanity_check > 1) {
        fprintf(stderr,"%s,%d: ARGHHH\n",__FILE__,__LINE__);
        exit(666);
      }
    }
  }
  return current_comm_ct;
}

/* compute_current_comp_ct() 
 *  Computes the time at which currently 
 *  RUNNING computation will complete. WARNING: this includes
 *  computations related to currently running communications. 
 */
static long compute_current_comp_ct(host_t *host)
{
  int i;
  int sanity_check;
  long current_comm_ct, current_comp_ct;


  /* Compute the time at which the (single) currently running 
     computation will complete */ 
  double current_comp_ct_double = (double)(global.current_date + host->time_to_prog_download);
  // fprintf(stderr,"\t\tcurrent_comp_ct_double=%f\n",current_comp_ct_double);
  sanity_check = 0;
  for (i=0; i < global.config->num_tasks; i++) {
    task_t *task = global.config->tasks[i];
    /* If the task isn't activated, forget it */
    if (!task->activated)
      continue;
    /* If task isn't scheduled on host, forget it */
    if (task->assigned_host != host)
      continue;

    if ((task->remaining_work_fraction < 1.0) && (task->remaining_work_fraction > EPSILON)) {
      // fprintf(stderr,"\t\t--> %.20f\n",task->remaining_work_fraction);
      current_comp_ct_double += (double)(host->w) * task->remaining_work_fraction;
      sanity_check++;
      /* Sanity check: we shoudn't do this more than once!! */
      if (sanity_check > 1) {
        fprintf(stderr,"%s,%d: ARGHHH (host %d)\n",__FILE__,__LINE__,host->id);
        exit(666);
      }
    }
  }

  /* If there is a pending communication, then we MUST do the
     corresponding computation after the current pending one has completed */
  //fprintf(stderr,"\t\tcurrent_comp_ct_double=%.40f\n",current_comp_ct_double);
  current_comm_ct = compute_current_comm_ct(host);
  //fprintf(stderr,"\t\tcurrent_comp_ct=%ld\n",current_comp_ct);
  if (current_comm_ct > global.current_date) {
    current_comp_ct += host->w;
  }
  current_comp_ct = (long) ceil(current_comp_ct_double);

  //fprintf(stderr,"\t\tcurrent_comp_ct=%ld\n",current_comp_ct);

  return current_comp_ct;
}

/* will_host_communicate_in_the_future() 
 *  Is there some assigned task on the host that hasn't started commmunication
 */
static char will_host_communicate_in_the_future(host_t *host)
{
  int i;

  for (i=0; i < global.config->num_tasks; i++) {
    task_t *task = global.config->tasks[i];
    /* If the task isn't activated or not on host, forget it */
    if ((!task->activated) || (task->assigned_host != host))
      continue;
    if (task->time_to_data_download > 0) 
      return 1;
  }
  return 0;
}

/* compute_contention_factor()
 */
static double compute_contention_factor(void)
{
  int i;
  int num_communicating_hosts;

  num_communicating_hosts = 0;
  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = global.platform.hosts[i];
    // If host is down, forget it
    if (STATE(host) == HOST_DOWN) 
      continue;
    if (will_host_communicate_in_the_future(host))
      num_communicating_hosts++;
  }
  return ((double)num_communicating_hosts / (double)global.platform.max_num_comms);
}

/* compute_expected_ct()
*/
static double compute_expected_ct(host_t *host, long ct)
{
  double ect;
  long W;

  W = (ct - global.current_date);

  ect = global.current_date + W + (W-1) * 
       ( host->chain.matrix[HOST_UP][HOST_RECLAIMED] *
         host->chain.matrix[HOST_RECLAIMED][HOST_UP] /
         (1.0 - host->chain.matrix[HOST_RECLAIMED][HOST_RECLAIMED]) ) /
       ( host->chain.matrix[HOST_UP][HOST_UP] * 
         (1.0 - host->chain.matrix[HOST_RECLAIMED][HOST_RECLAIMED]) +
         host->chain.matrix[HOST_UP][HOST_RECLAIMED] *
         host->chain.matrix[HOST_RECLAIMED][HOST_UP]);

  return ect;
}

/* compute_expected_rate()
*/
static double compute_expected_rate(host_t *host, double W, 
                                    char conditional, char naive, char scaled)
{
  double expected_rate = 0.0;

  if (conditional == CONDITIONAL) {
    expected_rate = compute_erate(
                     host->chain.matrix[HOST_UP][HOST_UP],
                     host->chain.matrix[HOST_UP][HOST_RECLAIMED],
                     host->chain.matrix[HOST_RECLAIMED][HOST_UP],
                     host->chain.matrix[HOST_RECLAIMED][HOST_RECLAIMED],
		     W, host->Pfanny, 
                     1+num_tasks_not_yet_completed_on_host(host),
                     host->id);
  } else if ((conditional == NOT_CONDITIONAL) && (naive == NAIVE)) {
    expected_rate = pow(host->Pfanny, W-1) * 
           compute_erate(host->chain.matrix[HOST_UP][HOST_UP],
                         host->chain.matrix[HOST_UP][HOST_RECLAIMED],
                         host->chain.matrix[HOST_RECLAIMED][HOST_UP],
                         host->chain.matrix[HOST_RECLAIMED][HOST_RECLAIMED],
		         W, host->Pfanny, 
                         1+num_tasks_not_yet_completed_on_host(host),
                         host->id);
    
  } else if ((conditional == NOT_CONDITIONAL) && (naive == NOT_NAIVE)) {
    expected_rate = 0;
    int i;
    double n = num_tasks_not_yet_started_on_host(host) + 1;
    for (i=0; i < n-1; i++) {
      long W_this_task = compute_MCT_indep_task_ct(host, scaled, i);
      long W_next_task = compute_MCT_indep_task_ct(host, scaled, i+1);

      expected_rate += pow(host->Pfanny, MAX(W_this_task-1,1)) * 
                (1.0 - pow(host->Pfanny, MAX(W_next_task-1,1))) * 
                compute_erate(host->chain.matrix[HOST_UP][HOST_UP],
                              host->chain.matrix[HOST_UP][HOST_RECLAIMED],
                              host->chain.matrix[HOST_RECLAIMED][HOST_UP],
                              host->chain.matrix[HOST_RECLAIMED][HOST_RECLAIMED],
                              W_this_task, host->Pfanny, i+1, host->id);
    }

  } else {
    fprintf(stderr,"compute_expected_rate(): invalid parameters\n");
    exit(1);
  }
  return expected_rate;
}

static double compute_erate(double Puu, double Pur, double Pru, double Prr, 
                     double W, double Pfanny, int n, int hostid)
{
  double erate;
  double term_1, term_2, term_3;
  double integral_1, integral_2, integral_3;

  /* Compute the first integral */
  {
    double abserr;
    size_t limit=10000;
    struct params_function_rate_integral parameters;

    gsl_integration_workspace *workspace = gsl_integration_workspace_alloc(limit);
    gsl_function F;
    F.function = &function_rate_integral_1;
    parameters.a = (Pur * Pru) / (Puu * Prr);
    parameters.b = W-1;
    F.params = &(parameters);
    if (gsl_integration_qag(&F,0,Prr,
                            0.0001,0.0001, limit, GSL_INTEG_GAUSS61, workspace,
                            &integral_1, &abserr)) {
      fprintf(stderr,"GSL integration error (integral #1) - returning 0.0\n");
      return 0.0;
    }
    gsl_integration_workspace_free(workspace);
  }
//  fprintf(stderr,"\tintegral #1: %.40lf\n",integral_1);

  term_1 = (pow(Puu / Prr, W-1) / Prr) * integral_1;
  //fprintf(stderr,"\tfactor_1= %.40lf\n",pow(Puu / Prr, W-1) / Prr);
  //fprintf(stderr,"\tterm_1=   %.40lf\n",term_1);

  /* Compute the second integral */
  {
    double abserr;
    size_t limit=100000;
    struct params_function_rate_integral parameters;

    gsl_integration_workspace *workspace = gsl_integration_workspace_alloc(limit);
    gsl_function F;
    F.function = &function_rate_integral_2;
    parameters.a = Puu * Puu / (Pur * Pru);
    parameters.b = W-1;
    F.params = &(parameters);

    if (gsl_integration_qag(&F,0,(Pur * Pru) / Puu,
                            0.00000000001,0.00000000001, limit, GSL_INTEG_GAUSS61, workspace,
                            &integral_2, &abserr)) {
      fprintf(stderr,"GSL integration error (integral #2) - returning 0.0\n");
      return 0.0;
    }
    gsl_integration_workspace_free(workspace);
  }
  //fprintf(stderr,"\n\tintegral #2: %.40lf\n",integral_2);


  term_2 = (Puu / (Pru * Pur)) * integral_2;
  //fprintf(stderr,"\tfactor_2= %.40lf\n",(Puu / (Pru * Pur)));
  //fprintf(stderr,"\tterm_2=   %.40lf\n",term_2);


  /* Compute the third integral */
  {
    double abserr;
    size_t limit=100000;
    struct params_function_rate_integral parameters;

    gsl_integration_workspace *workspace = gsl_integration_workspace_alloc(limit);
    gsl_function F;
    F.function = &function_rate_integral_3;
    parameters.a = Puu;
    parameters.b = Pur * Pru ;
    parameters.c = W-1;
    F.params = &(parameters);

    if (gsl_integration_qag(&F,0,1,
                            0.00000000001,0.00000000001, limit, GSL_INTEG_GAUSS61, workspace,
                            &integral_3, &abserr)) {
      fprintf(stderr,"GSL integration error (integral #3) - returning 0.0\n");
      return 0.0;
    }
    gsl_integration_workspace_free(workspace);
  }
  //fprintf(stderr,"\n\tintegral #3: %.40lf\n",integral_3);

  term_3 = integral_3;
  //fprintf(stderr,"\tterm_3=   %.40lf\n",term_3);

  double sum_of_terms = (term_1 + term_2 - term_3);

  erate = n * sum_of_terms / (pow(Pfanny,W-1));

//  fprintf(stderr,"host#%d: W=%.1f, sum of terms = %.3f, ERate=%.3f\n",
//                 hostid,W,sum_of_terms,erate);

  if (erate != erate) {
    fprintf(stderr,"NaN\n"); exit(1);
  }

//  fprintf(stderr,"rate=%.40f\n",erate);

  return erate;
}

/* function_rate_integral_1()
 *
 * A function of the form
 *   (x + a * x^2 / (1-x))^b
 *  
 */
static double function_rate_integral_1(double x, void *p) {
  struct params_function_rate_integral *parameters = (struct params_function_rate_integral*)p;
  double a = parameters->a;
  double b = parameters->b;

  return pow((x + a * x * x / (1 - x)), b);
}

/* function_rate_integral_2()
 *
 * A function of the form  (x * (1 + x) * a)^b
 *  
 */
static double function_rate_integral_2(double x, void *p) {
  struct params_function_rate_integral *parameters = (struct params_function_rate_integral*)p;
  double a = parameters->a;
  double b = parameters->b;

  return pow(x * (1 + x) * a, b);
}

/* function_rate_integral_3()
 *
 * A function of the form  ( a*x + b*x^2 )^c
 *  
 */
static double function_rate_integral_3(double x, void *p) {
  struct params_function_rate_integral *parameters = (struct params_function_rate_integral*)p;
  double a = parameters->a;
  double b = parameters->b;
  double c = parameters->c;

  return pow(a*x + b*x*x, c);
}




/* create_clones() */
static void create_clones(void)
{
  int task_index;
  int num_allowed_clones;

  if (global.max_num_clones == 0)
    return;

  num_allowed_clones = how_many_allowed_clones();

  while (num_allowed_clones--) {

    if ((task_index = pick_task_to_clone()) == -1)
      break;

    create_task_clone(task_index);
  }

  return;
}

/* how_many_allowed_clones()
 *  Could be optimize to return a number of allowed clones
 */
static int how_many_allowed_clones(void)
{
  int num_up_hosts;
  int num_activated_tasks;
  int num_will_communicate_tasks;
  int i;

  num_up_hosts = 0;
  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = global.platform.hosts[i];
    if (STATE(host) == HOST_UP)
      num_up_hosts++;
  }
 
  num_activated_tasks = 0; 
  for (i=0; i < global.config->num_tasks; i++) {
    task_t *task = global.config->tasks[i];
    if ((task->activated) && (!task->done))
      num_activated_tasks++;
  }

  num_will_communicate_tasks = 0; 
  for (i=0; i < global.config->num_tasks; i++) {
    task_t *task = global.config->tasks[i];
    if ((task->activated) && (!task->done) && (task->time_to_data_download > 0))
      num_will_communicate_tasks++;
  }

  return MAX(0, MIN(num_up_hosts - num_activated_tasks, 
                    global.platform.max_num_comms - num_will_communicate_tasks));
}

/* pick_task_to_clone()
 *
 *  - Create a list of not-done tasks that have the fewest clones
 *  - Select the task among these that has the lowest Pfanny ^ W
 */
static int pick_task_to_clone(void)
{
  int i,smallest_number_of_clones;


  /* Compute the smallest number of clones among tasks */
  smallest_number_of_clones = 1 + global.max_num_clones;
  for (i=0; i < global.application.num_tasks; i++) {
    int number_of_clones = 0;

    if (global.config->tasks[i * global.platform.num_hosts]->done)
      continue;

    number_of_clones = get_num_of_clones(i);

    if (number_of_clones < smallest_number_of_clones) {
      smallest_number_of_clones = number_of_clones;
    }
  }

  /* If cloned-out, then forget it */
  if (smallest_number_of_clones == 1 + global.max_num_clones)
    return -1;

  /* Make a list of potentially clonable tasks */
  int *clonable = NULL;
  int num_clonable = 0;
  for (i=0; i < global.application.num_tasks; i++) {
    if (get_num_of_clones(i) == smallest_number_of_clones) {
      clonable = (int *)calloc(num_clonable+1,sizeof(int)); 
      clonable[num_clonable] = i;
      num_clonable++;
    }
  }

  /* Pick the one with the largest Pfanny ^ W */
  int picked = 1;
  double picked_proba_success = compute_proba_success(clonable[1]);
  for (i=1; i < num_clonable; i++) {
    double proba_success = compute_proba_success(clonable[i]);
    if (proba_success > picked_proba_success) {
      picked = i;
      picked_proba_success = proba_success; 
    }
  }
  free(clonable);

  return picked;
}

/* create_task_clone */
static void create_task_clone(int task_index)
{
  int j;
  
  for (j=1; j < global.platform.num_hosts; j++) {
    task_t *potential_clone = 
             global.config->tasks[task_index * global.platform.num_hosts + j];
    if (!potential_clone->activated) {
      potential_clone->activated = TRUE;
      break;
    }
  }
  return;
}

/* get_num_of_clones() */
static int get_num_of_clones(int task_index)
{
  int j,number_of_clones = 0;

  for (j=0; j < global.platform.num_hosts; j++) {
    if (global.config->tasks[task_index * global.platform.num_hosts + j]->activated)
      number_of_clones++;
  } 
  return number_of_clones;
}

/* get_hosts_with_clone() 
 *  Returns a NULL-terminated list 
 */
static host_t **get_hosts_with_clone(task_t *task)
{
  int i,j;
  int num_hosts = 0;
  host_t **host_list = NULL;

  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = global.platform.hosts[i];

    for (j=0; j < global.platform.num_hosts; j++) {
      task_t *instance = global.config->tasks[task->id * global.platform.num_hosts + j];
      if (instance->activated && (!task->done) && (task->assigned_host == host)) {
        host_list = (host_t **)realloc(host_list, (num_hosts+1)*sizeof(host_t*));
        host_list[num_hosts] = host;
        num_hosts++;
      }
    }
  }

  // Null-terminate the list
  host_list = (host_t **)realloc(host_list, (num_hosts+1)*sizeof(host_t*));
  host_list[num_hosts] = NULL;
  
  return host_list;
}

/* is_host_in_list() */
static int is_host_in_list(host_t *host, host_t **dont_pick) { 

  while (*dont_pick) {
    if (*dont_pick == host)
      return TRUE;
    dont_pick++;
  }  
  return FALSE; 
}

/* compute_proba_success() 
 * Pfanny ^ W
 */
static double compute_proba_success(int task_index)
{
  int i,j;

   /* Does not take into account overlap between computation and communication */

  double proba = 1.0;
  for (j=0; j < global.platform.num_hosts; j++) {
    task_t *task = global.config->tasks[task_index * global.platform.num_hosts + j];
    host_t *host = task->assigned_host;

    if ((!host) || (!task->activated))
      continue;

    /* Compute W */
    double W = 0.0;
    for (i=0; i < global.config->num_tasks; i++) {
      if (i == task->id)
        continue;

      task_t *other_task = global.config->tasks[i];
      if (other_task->assigned_host != host) 
        continue;
      if (!other_task->activated)
        continue;

      /* Compute the number of communication time-step for all non-clone
       * tasks on this host, since non-clones are executed in priority */
      if (task->time_to_data_download == global.application.Tdata) {
        W += other_task->time_to_data_download;
      } else if (task->time_to_data_download > 0) {
        if (other_task->time_to_data_download < global.application.Tdata) {
          W += other_task->time_to_data_download;
        }
      }

      /* Compute the number of computation time-step for all non-clone
       * tasks on this host, since non-clones are executed in priority */
      if (task->remaining_work_fraction == 1.0) {
        W += host->w * other_task->remaining_work_fraction;
      } else if (task->remaining_work_fraction > EPSILON) {
        if (other_task->remaining_work_fraction < 1.0) {
          W += host->w * other_task->remaining_work_fraction;
        }
      }
    }

    W += task->time_to_data_download;
    W += host->w * task->remaining_work_fraction;

    /* Update the proba */
    proba *= (1.0 - pow(host->Pfanny, (double)W));
  }
  return (1.0 - proba);
}


/*********************************************/
/***************** COUPLED *******************/
/*********************************************/

/*schedule_coupled_proactif()*/
static int schedule_coupled_proactif(config_t *config, host_selection_type_t host_selection_type, const char *criterion)
{
  double current_weight,weight;
  task_t *current_config;
  int i;
  host_t *current_hosts;

  /*current configuration*/
  schedule_coupled(config, host_selection_type);
  
  if (!strcmp(criterion,"PP")) {
    current_weight=coupled_success_proba();
  } else if (!strcmp(criterion,"PE")) {
    current_weight=-coupled_exp_completion_time();
  } else if (!strcmp(criterion,"PR")) {
    current_weight=coupled_success_proba()/coupled_exp_completion_time();
  }else if (!strcmp(criterion,"PY")) {
    current_weight=coupled_success_proba()/(coupled_exp_completion_time()+(double)(global.current_date-global.application.date_of_current_iteration));
  }

  /*copy and remove the current configuration*/
  current_config=malloc(global.config->num_tasks*sizeof(task_t));
  current_hosts=malloc(global.platform.num_hosts*sizeof(host_t));
  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = global.platform.hosts[i];
    current_hosts[i]=*host;
  }

  for( i=0;i<global.config->num_tasks;i++)
  {
    task_t *task=global.config->tasks[i];
    current_config[i]=*task;
    if (task->assigned_host) {
      unassign_task(task);
    }
  }

  /*new configuration*/
  schedule_coupled(config, host_selection_type);

  if (!strcmp(criterion,"PP")) {
    weight=coupled_success_proba();
  } else if (!strcmp(criterion,"PE")) {
    weight=-coupled_exp_completion_time();
  } else if (!strcmp(criterion,"PR")) {
    weight=coupled_success_proba()/coupled_exp_completion_time();
  }else if (!strcmp(criterion,"PY")) {
    weight=coupled_success_proba()/(coupled_exp_completion_time()+(double)(global.current_date-global.application.date_of_current_iteration));
  }

  /*comparison*/
  if(weight<=current_weight){
    for(i=0;i<global.config->num_tasks;i++)
    {
      *global.config->tasks[i]=current_config[i];
    }
    for (i=0; i < global.platform.num_hosts; i++) {
      *global.platform.hosts[i]=current_hosts[i];
    }
  }

  free(current_config);
  free(current_hosts);

  return 0;
}


static int schedule_coupled(config_t *config,
                            host_selection_type_t host_selection_type)
{
  int i,j;

  /* Reset all the host counters that will be used to track which procs
     were "picked" first for tasks, so as to prioritize communications
     accordingly (proc picked first -> new communication to it has priority */
  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = global.platform.hosts[i];
    host->picking_order = global.config->num_tasks; // worst ranking + 1
  }

  /* Create clones */
  create_clones();

  /* Assign all tasks that are unassigned, according to some criterion */
  /* Consider all original tasks first, then all 1st clones, etc. */
  for (j=0; j < global.platform.num_hosts; j++) {
    for (i=0; i < global.application.num_tasks; i++) {
      task_t *task =config->tasks[i * global.platform.num_hosts + j];
      host_t *host = NULL;

      /* If the task is not activated, forget it */
      if (!task->activated)
        continue;

      /* If the task is assigned or done, leave it where it is */
      if ((task->assigned_host) || (task->done))
        continue;

      /* Build the list of hosts to avoid because a clone of
         the task's already scheduled/running on that host */
      host_t **dont_pick = get_hosts_with_clone(task);

      /* pick a host */
      switch(host_selection_type) {
	  
	case RANDOM:
          host = pick_random_host_up("none",dont_pick);
          break;

        case WEIGHTED_RANDOM_Puu:
          host = pick_random_host_up("Puu",dont_pick);
          break;

        case WEIGHTED_RANDOM_Puuw:
          host = pick_random_host_up("Puuw",dont_pick);
          break;

        case WEIGHTED_RANDOM_Pfanny:
          host = pick_random_host_up("Pfanny",dont_pick);
          break;

        case WEIGHTED_RANDOM_Pfannyw:
          host = pick_random_host_up("Pfannyw",dont_pick);
          break;

        case WEIGHTED_RANDOM_PIu:
          host = pick_random_host_up("PIu",dont_pick);
          break;

        case WEIGHTED_RANDOM_PIuw:
          host = pick_random_host_up("PIuw",dont_pick);
          break;

        case WEIGHTED_RANDOM_PIuf:
          host = pick_random_host_up("PIuf",dont_pick);
          break;

        case WEIGHTED_RANDOM_PIufw:
          host = pick_random_host_up("PIufw",dont_pick);
          break;

        case MCT_v1:
          host = pick_MCT_host_up(NOT_SCALED,NOT_EXPECTED,dont_pick);
          break;

        case MCT_v2:
          host = pick_MCT_host_up(SCALED,NOT_EXPECTED,dont_pick);
          break;

        case LW_v1:
          host = pick_LW_host_up(NOT_SCALED,NOT_EXPECTED,dont_pick);
          break;

        case LW_v2:
          host = pick_LW_host_up(SCALED,NOT_EXPECTED,dont_pick);
          break;

        case UD_v1:
          host = pick_UD_host_up(NOT_SCALED,NOT_EXPECTED,dont_pick);
          break;

        case UD_v2:
          host = pick_UD_host_up(SCALED,NOT_EXPECTED,dont_pick);
          break;

        case EMCT_v1:
          host = pick_MCT_host_up(NOT_SCALED,EXPECTED,dont_pick);
          break;

        case EMCT_v2:
          host = pick_MCT_host_up(SCALED,EXPECTED,dont_pick);
          break;

        case ELW_v1:
          host = pick_LW_host_up(NOT_SCALED,EXPECTED,dont_pick);
          break;

        case ELW_v2:
          host = pick_LW_host_up(SCALED,EXPECTED,dont_pick);
          break;

        case EUD_v1:
          host = pick_UD_host_up(NOT_SCALED,EXPECTED,dont_pick);
          break;

        case EUD_v2:
          host = pick_UD_host_up(SCALED,EXPECTED,dont_pick);
	  break;
	
	case IE:
	  host = pick_inc_host_up("IE",dont_pick,task);
	  break;
	
	case IP:
	  host = pick_inc_host_up("IP",dont_pick,task);
	  break;
	
	case IR:
	  host = pick_inc_host_up("IR",dont_pick,task);
	  break;

	case IY:
	  host = pick_inc_host_up("IY",dont_pick,task);
	  break;
	
	default:
	  fprintf(stderr,"heuristic indep in model coupled\n");
	  exit(1);

      }
      free(dont_pick);

      /* Assign the task to host if a host was found */
	  if (host) {
	    if (host->picking_order > i) {
	      host->picking_order = i;
	    }
	    assign_task_to_host(task,host);
	    if(host->data_download_time>=host->num_assigned_tasks*global.application.Tdata)
	    {
	      task->time_to_data_download=0;
	    } else if(host->data_download_time>=(host->num_assigned_tasks-1)*global.application.Tdata){
	      task->time_to_data_download-=host->data_download_time-(host->num_assigned_tasks-1)*global.application.Tdata;}
	  }
    }
  }
  for (j=0; j < global.platform.num_hosts; j++) 
  {
    host_t *host=global.platform.hosts[j];
    if(host->data_download_time>host->num_assigned_tasks*global.application.Tdata)
    {host->data_download_time=host->num_assigned_tasks*global.application.Tdata;}
  }
  return 0;
}

/* schedule_coupled_random() */
int schedule_coupled_random(config_t *config)
{
  return schedule_coupled(config, RANDOM);
  // return schedule_coupled(config, TEST);
}

int schedule_coupled_random_PP(config_t *config)
{
  return schedule_coupled_proactif(config, RANDOM, "PP");
}

int schedule_coupled_random_PE(config_t *config)
{
  return schedule_coupled_proactif(config, RANDOM,"PE");
  // return schedule_coupled(config, TEST);
}

int schedule_coupled_random_PR(config_t *config)
{
  return schedule_coupled_proactif(config, RANDOM, "PR");
}


int schedule_coupled_random_PY(config_t *config)
{
  return schedule_coupled_proactif(config, RANDOM, "PY");
  // return schedule_coupled(config, TEST);
}

/* schedule_coupled_weighted_random_Puu() */
int schedule_coupled_weighted_random_Puu(config_t *config) {
  return schedule_coupled(config, WEIGHTED_RANDOM_Puu);
}

/* schedule_coupled_weighted_random_Puu_PP() */
int schedule_coupled_weighted_random_Puu_PP(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Puu,"PP");
}

/* schedule_coupled_weighted_random_Puu_PE() */
int schedule_coupled_weighted_random_Puu_PE(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Puu,"PE");
}

/* schedule_coupled_weighted_random_Puu_PR() */
int schedule_coupled_weighted_random_Puu_PR(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Puu,"PR");
}

/* schedule_coupled_weighted_random_Puu_PY() */
int schedule_coupled_weighted_random_Puu_PY(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Puu,"PY");
}

/* schedule_coupled_weighted_random_Puuw() */
int schedule_coupled_weighted_random_Puuw(config_t *config) {
  return schedule_coupled(config, WEIGHTED_RANDOM_Puuw);
}

/* schedule_coupled_weighted_random_Puuw_PP() */
int schedule_coupled_weighted_random_Puuw_PP(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Puuw,"PP");
}

/* schedule_coupled_weighted_random_Puuw_PE() */
int schedule_coupled_weighted_random_Puuw_PE(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Puuw,"PE");
}

/* schedule_coupled_weighted_random_Puuw_PR() */
int schedule_coupled_weighted_random_Puuw_PR(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Puuw,"PR");
}

/* schedule_coupled_weighted_random_Puuw_PY() */
int schedule_coupled_weighted_random_Puuw_PY(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Puuw,"PY");
}

/* schedule_coupled_weighted_random_Pfanny() */
int schedule_coupled_weighted_random_Pfanny(config_t *config) {
  return schedule_coupled(config, WEIGHTED_RANDOM_Pfanny);
}

/* schedule_coupled_weighted_random_Pfanny_PP() */
int schedule_coupled_weighted_random_Pfanny_PP(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Pfanny,"PP");
}

/* schedule_coupled_weighted_random_Pfanny_PE() */
int schedule_coupled_weighted_random_Pfanny_PE(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Pfanny,"PE");
}

/* schedule_coupled_weighted_random_Pfanny_PR() */
int schedule_coupled_weighted_random_Pfanny_PR(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Pfanny,"PR");
}
  
/* schedule_coupled_weighted_random_Pfanny_PY() */
int schedule_coupled_weighted_random_Pfanny_PY(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Pfanny,"PY");
}

/* schedule_coupled_weighted_random_Pfannyw() */
int schedule_coupled_weighted_random_Pfannyw(config_t *config) {
  return schedule_coupled(config, WEIGHTED_RANDOM_Pfannyw);
}

/* schedule_coupled_weighted_random_Pfannyw_PP() */
int schedule_coupled_weighted_random_Pfannyw_PP(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Pfannyw,"PP");
}

/* schedule_coupled_weighted_random_Pfannyw_PE() */
int schedule_coupled_weighted_random_Pfannyw_PE(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Pfannyw,"PE");
}

/* schedule_coupled_weighted_random_Pfannyw_PR() */
int schedule_coupled_weighted_random_Pfannyw_PR(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Pfannyw,"PR");
}

/* schedule_coupled_weighted_random_Pfannyw_PY() */
int schedule_coupled_weighted_random_Pfannyw_PY(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_Pfannyw,"PY");
}

/* schedule_coupled_weighted_random_PIu() */
int schedule_coupled_weighted_random_PIu(config_t *config) {
  return schedule_coupled(config, WEIGHTED_RANDOM_PIu);
}

/* schedule_coupled_weighted_random_PIu_PP() */
int schedule_coupled_weighted_random_PIu_PP(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIu,"PP");
}

/* schedule_coupled_weighted_random_PIu_PE() */
int schedule_coupled_weighted_random_PIu_PE(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIu,"PE");
}

/* schedule_coupled_weighted_random_PIu_PR() */
int schedule_coupled_weighted_random_PIu_PR(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIu,"PR");
}

/* schedule_coupled_weighted_random_PIu_PY() */
int schedule_coupled_weighted_random_PIu_PY(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIu,"PY");
}

/* schedule_coupled_weighted_random_PIuw() */
int schedule_coupled_weighted_random_PIuw(config_t *config) {
  return schedule_coupled(config, WEIGHTED_RANDOM_PIuw);
}

/* schedule_coupled_weighted_random_PIuw_PP() */
int schedule_coupled_weighted_random_PIuw_PP(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIuw,"PP");
}

/* schedule_coupled_weighted_random_PIuw_PE() */
int schedule_coupled_weighted_random_PIuw_PE(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIuw,"PE");
} 

/* schedule_coupled_weighted_random_PIuw_PR() */
int schedule_coupled_weighted_random_PIuw_PR(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIuw,"PR");
}

/* schedule_coupled_weighted_random_PIuw_PY() */
int schedule_coupled_weighted_random_PIuw_PY(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIuw,"PY");
}

/* schedule_coupled_weighted_random_PIuf() */
int schedule_coupled_weighted_random_PIuf(config_t *config) {
  return schedule_coupled(config, WEIGHTED_RANDOM_PIuf);
}

/* schedule_coupled_weighted_random_PIuf_PP() */
int schedule_coupled_weighted_random_PIuf_PP(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIuf,"PP");
}

/* schedule_coupled_weighted_random_PIuf_PE() */
int schedule_coupled_weighted_random_PIuf_PE(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIuf,"PE");
} 

/* schedule_coupled_weighted_random_PIuf_PR() */
int schedule_coupled_weighted_random_PIuf_PR(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIuf,"PR");
}

/* schedule_coupled_weighted_random_PIuf_PY() */
int schedule_coupled_weighted_random_PIuf_PY(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIuf,"PY");
}

/* schedule_coupled_weighted_random_PIufw() */
int schedule_coupled_weighted_random_PIufw(config_t *config) {
  return schedule_coupled(config, WEIGHTED_RANDOM_PIufw);
}

/* schedule_coupled_weighted_random_PIufw_PP() */
int schedule_coupled_weighted_random_PIufw_PP(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIufw,"PP");
} 

/* schedule_coupled_weighted_random_PIufw_PE() */
int schedule_coupled_weighted_random_PIufw_PE(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIufw,"PE");
}   

/* schedule_coupled_weighted_random_PIufw_PR() */
int schedule_coupled_weighted_random_PIufw_PR(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIufw,"PR");
}     

/* schedule_coupled_weighted_random_PIufw_PY() */
int schedule_coupled_weighted_random_PIufw_PY(config_t *config) {
  return schedule_coupled_proactif(config, WEIGHTED_RANDOM_PIufw,"PY");
}

/* schedule_MCT_coupled_v1() */
int schedule_coupled_MCT_v1(config_t *config) {
  return schedule_coupled(config, MCT_v1);
}

/* schedule_MCT_coupled_v1_PP() */
int schedule_coupled_MCT_v1_PP(config_t *config) {
  return schedule_coupled_proactif(config, MCT_v1,"PP");
}

/* schedule_MCT_coupled_v1_PE() */
int schedule_coupled_MCT_v1_PE(config_t *config) {
  return schedule_coupled_proactif(config, MCT_v1,"PE");
}

/* schedule_MCT_coupled_v1_PR() */
int schedule_coupled_MCT_v1_PR(config_t *config) {
  return schedule_coupled_proactif(config, MCT_v1,"PR");
}

/* schedule_MCT_coupled_v1_PY() */
int schedule_coupled_MCT_v1_PY(config_t *config) {
  return schedule_coupled_proactif(config, MCT_v1,"PY");
}

/* schedule_MCT_coupled_v2() */
int schedule_coupled_MCT_v2(config_t *config) {
  return schedule_coupled(config, MCT_v2);
}

/* schedule_MCT_coupled_v2_PP() */
int schedule_coupled_MCT_v2_PP(config_t *config) {
  return schedule_coupled_proactif(config, MCT_v2,"PP");
}

/* schedule_MCT_coupled_v2_PE() */
int schedule_coupled_MCT_v2_PE(config_t *config) {
  return schedule_coupled_proactif(config, MCT_v2,"PE");
}

/* schedule_MCT_coupled_v2_PR() */
int schedule_coupled_MCT_v2_PR(config_t *config) {
  return schedule_coupled_proactif(config, MCT_v2,"PR");
}

/* schedule_MCT_coupled_v2_PY() */
int schedule_coupled_MCT_v2_PY(config_t *config) {
  return schedule_coupled_proactif(config, MCT_v2,"PY");
}

/* schedule_UD_coupled_v1() */
int schedule_coupled_UD_v1(config_t *config) {
  return schedule_coupled(config, UD_v1);
}

/* schedule_UD_coupled_v1_PP() */
int schedule_coupled_UD_v1_PP(config_t *config) {
  return schedule_coupled_proactif(config, UD_v1,"PP");
}

/* schedule_UD_coupled_v1_PE() */
int schedule_coupled_UD_v1_PE(config_t *config) {
  return schedule_coupled_proactif(config, UD_v1,"PE");
}

/* schedule_UD_coupled_v1_PR() */
int schedule_coupled_UD_v1_PR(config_t *config) {
  return schedule_coupled_proactif(config, UD_v1,"PR");
}

/* schedule_UD_coupled_v1_PY() */
int schedule_coupled_UD_v1_PY(config_t *config) {
  return schedule_coupled_proactif(config, UD_v1,"PY");
}

/* schedule_UD_coupled_v2() */
int schedule_coupled_UD_v2(config_t *config) {
  return schedule_coupled(config, UD_v2);
}

/* schedule_UD_coupled_v2_PP() */
int schedule_coupled_UD_v2_PP(config_t *config) {
  return schedule_coupled_proactif(config, UD_v2,"PP");
}

/* schedule_UD_coupled_v2_PE() */
int schedule_coupled_UD_v2_PE(config_t *config) {
  return schedule_coupled_proactif(config, UD_v2,"PE");
}

/* schedule_UD_coupled_v2_PR() */
int schedule_coupled_UD_v2_PR(config_t *config) {
  return schedule_coupled_proactif(config, UD_v2,"PR");
}
        
/* schedule_UD_coupled_v2_PY() */
int schedule_coupled_UD_v2_PY(config_t *config) {
  return schedule_coupled_proactif(config, UD_v2,"PY");
}

/* schedule_LW_coupled_v1() */
int schedule_coupled_LW_v1(config_t *config) {
  return schedule_coupled(config, LW_v1);
}

/* schedule_LW_coupled_v1_PP() */
int schedule_coupled_LW_v1_PP(config_t *config) {
  return schedule_coupled_proactif(config, LW_v1,"PP");
}

/* schedule_LW_coupled_v1_PE() */
int schedule_coupled_LW_v1_PE(config_t *config) {
  return schedule_coupled_proactif(config, LW_v1,"PE");
}

/* schedule_LW_coupled_v1_PR() */
int schedule_coupled_LW_v1_PR(config_t *config) {
  return schedule_coupled_proactif(config, LW_v1,"PR");
}

/* schedule_LW_coupled_v1_PY() */
int schedule_coupled_LW_v1_PY(config_t *config) {
  return schedule_coupled_proactif(config, LW_v1,"PY");
}

/* schedule_LW_coupled_v2() */
int schedule_coupled_LW_v2(config_t *config) {
  return schedule_coupled(config, LW_v2);
}

/* schedule_LW_coupled_v2_PP() */
int schedule_coupled_LW_v2_PP(config_t *config) {
  return schedule_coupled_proactif(config, LW_v1,"PP");
}

/* schedule_LW_coupled_v2_PE() */
int schedule_coupled_LW_v2_PE(config_t *config) {
  return schedule_coupled_proactif(config, LW_v1,"PE");
}

/* schedule_LW_coupled_v2_PR() */
int schedule_coupled_LW_v2_PR(config_t *config) {
  return schedule_coupled_proactif(config, LW_v1,"PR");
}

/* schedule_LW_coupled_v2_PY() */
int schedule_coupled_LW_v2_PY(config_t *config) {
  return schedule_coupled_proactif(config, LW_v1,"PY");
}

/* schedule_EMCT_coupled_v1() */
int schedule_coupled_EMCT_v1(config_t *config) {
  return schedule_coupled(config, EMCT_v1);
}

/* schedule_EMCT_coupled_v1_PP() */
int schedule_coupled_EMCT_v1_PP(config_t *config) {
  return schedule_coupled_proactif(config, EMCT_v1,"PP");
}

/* schedule_EMCT_coupled_v1_PE() */
int schedule_coupled_EMCT_v1_PE(config_t *config) {
  return schedule_coupled_proactif(config, EMCT_v1,"PE");
}

/* schedule_EMCT_coupled_v1_PR() */
int schedule_coupled_EMCT_v1_PR(config_t *config) {
  return schedule_coupled_proactif(config, EMCT_v1,"PR");
}

/* schedule_EMCT_coupled_v1_PY() */
int schedule_coupled_EMCT_v1_PY(config_t *config) {
  return schedule_coupled_proactif(config, EMCT_v1,"PY");
}

/* schedule_EMCT_coupled_v2() */
int schedule_coupled_EMCT_v2(config_t *config) {
  return schedule_coupled(config, EMCT_v2);
}

/* schedule_EMCT_coupled_v2_PP() */
int schedule_coupled_EMCT_v2_PP(config_t *config) {
  return schedule_coupled_proactif(config, EMCT_v2,"PP");
}

/* schedule_EMCT_coupled_v2_PE() */
int schedule_coupled_EMCT_v2_PE(config_t *config) {
  return schedule_coupled_proactif(config, EMCT_v2,"PE");
}

/* schedule_EMCT_coupled_v2_PR() */
int schedule_coupled_EMCT_v2_PR(config_t *config) {
  return schedule_coupled_proactif(config, EMCT_v2,"PR");
}

/* schedule_EMCT_coupled_v2_PY() */
int schedule_coupled_EMCT_v2_PY(config_t *config) {
  return schedule_coupled_proactif(config, EMCT_v2,"PY");
}

/* schedule_EUD_coupled_v1() */
int schedule_coupled_EUD_v1(config_t *config) {
  return schedule_coupled(config, EUD_v1);
}

/* schedule_EUD_coupled_v1_PP() */
int schedule_coupled_EUD_v1_PP(config_t *config) {
  return schedule_coupled_proactif(config, EUD_v1,"PP");
}

/* schedule_EUD_coupled_v1_PE() */
int schedule_coupled_EUD_v1_PE(config_t *config) {
  return schedule_coupled_proactif(config, EUD_v1,"PE");
}

/* schedule_EUD_coupled_v1_PR() */
int schedule_coupled_EUD_v1_PR(config_t *config) {
  return schedule_coupled_proactif(config, EUD_v1,"PR");
}

/* schedule_EUD_coupled_v1_PY() */
int schedule_coupled_EUD_v1_PY(config_t *config) {
  return schedule_coupled_proactif(config, EUD_v1,"PY");
}

/* schedule_EUD_coupled_v2() */
int schedule_coupled_EUD_v2(config_t *config) {
  return schedule_coupled(config, EUD_v2);
}

/* schedule_EUD_coupled_v2_PP() */
int schedule_coupled_EUD_v2_PP(config_t *config) {
  return schedule_coupled_proactif(config, EUD_v2,"PP");
}

/* schedule_EUD_coupled_v2_PE() */
int schedule_coupled_EUD_v2_PE(config_t *config) {
  return schedule_coupled_proactif(config, EUD_v2,"PE");
}

/* schedule_EUD_coupled_v2_PR() */
int schedule_coupled_EUD_v2_PR(config_t *config) {
  return schedule_coupled_proactif(config, EUD_v2,"PR");
}

/* schedule_EUD_coupled_v2_PY() */
int schedule_coupled_EUD_v2_PY(config_t *config) {
  return schedule_coupled_proactif(config, EUD_v2,"PY");
}

/* schedule_ELW_coupled_v1() */
int schedule_coupled_ELW_v1(config_t *config) {
  return schedule_indep(config, ELW_v1);
}

/* schedule_ELW_coupled_v1_PP() */
int schedule_coupled_ELW_v1_PP(config_t *config) {
  return schedule_coupled_proactif(config, ELW_v1,"PP");
}

/* schedule_ELW_coupled_v1_PE() */
int schedule_coupled_ELW_v1_PE(config_t *config) {
  return schedule_coupled_proactif(config, ELW_v1,"PE");
}

/* schedule_ELW_coupled_v1_PR() */
int schedule_coupled_ELW_v1_PR(config_t *config) {
  return schedule_coupled_proactif(config, ELW_v1,"PR");
}

/* schedule_ELW_coupled_v1_PY() */
int schedule_coupled_ELW_v1_PY(config_t *config) {
  return schedule_coupled_proactif(config, ELW_v1,"PY");
}

/* schedule_ELW_coupled_v2() */
int schedule_coupled_ELW_v2(config_t *config) {
  return schedule_coupled(config, ELW_v2);
}

/* schedule_ELW_coupled_v2_PP() */
int schedule_coupled_ELW_v2_PP(config_t *config) {
  return schedule_coupled_proactif(config, ELW_v2,"PP");
} 
    
/* schedule_ELW_coupled_v2_PE() */
int schedule_coupled_ELW_v2_PE(config_t *config) { 
  return schedule_coupled_proactif(config, ELW_v2,"PE");
}   
        
/* schedule_ELW_coupled_v2_PR() */
int schedule_coupled_ELW_v2_PR(config_t *config) {
  return schedule_coupled_proactif(config, ELW_v2,"PR");
}     

/* schedule_ELW_coupled_v2_PY() */
int schedule_coupled_ELW_v2_PY(config_t *config) {
  return schedule_coupled_proactif(config, ELW_v2,"PY");
}       

/* schedule_coupled_IE() */
int schedule_coupled_IE(config_t *config) {
  return schedule_coupled(config, IE);
}

/* schedule_coupled_IE_PP() */
int schedule_coupled_IE_PP(config_t *config) {
  return schedule_coupled_proactif(config, IE,"PP");
}

/* schedule_coupled_IE_PE() */
int schedule_coupled_IE_PE(config_t *config) {
  return schedule_coupled_proactif(config, IE,"PE");
}

/* schedule_coupled_IE_PR() */
int schedule_coupled_IE_PR(config_t *config) {
  return schedule_coupled_proactif(config, IE,"PR");
}

/* schedule_coupled_IE_PY() */
int schedule_coupled_IE_PY(config_t *config) {
  return schedule_coupled_proactif(config, IE,"PY");
}

/* schedule_coupled_IP() */
int schedule_coupled_IP(config_t *config) {
  return schedule_coupled(config, IP);
}

/* schedule_coupled_IP_PP() */
int schedule_coupled_IP_PP(config_t *config) {
  return schedule_coupled_proactif(config, IP,"PP");
}

/* schedule_coupled_IP_PE() */
int schedule_coupled_IP_PE(config_t *config) {
  return schedule_coupled_proactif(config, IP,"PE");
}

/* schedule_coupled_IP_PR() */
int schedule_coupled_IP_PR(config_t *config) {
  return schedule_coupled_proactif(config, IP,"PR");
}

/* schedule_coupled_IP_PY() */
int schedule_coupled_IP_PY(config_t *config) {
  return schedule_coupled_proactif(config, IP,"PY");
}

/* schedule_coupled_IR() */
int schedule_coupled_IR(config_t *config) {
  return schedule_coupled(config, IR);
}

/* schedule_coupled_IR_PP() */
int schedule_coupled_IR_PP(config_t *config) {
  return schedule_coupled_proactif(config, IR,"PP");
}

/* schedule_coupled_IR_PE() */
int schedule_coupled_IR_PE(config_t *config) {
  return schedule_coupled_proactif(config, IR,"PE");
}

/* schedule_coupled_IR_PR() */
int schedule_coupled_IR_PR(config_t *config) {
  return schedule_coupled_proactif(config, IR,"PR");
}

/* schedule_coupled_IR_PY() */
int schedule_coupled_IR_PY(config_t *config) {
  return schedule_coupled_proactif(config, IR,"PY");
}

/* schedule_coupled_IY() */
int schedule_coupled_IY(config_t *config) {
  return schedule_coupled(config, IY);
}

/* schedule_coupled_IY_PP() */
int schedule_coupled_IY_PP(config_t *config) {
  return schedule_coupled_proactif(config, IY,"PP");
} 

/* schedule_coupled_IY_PE() */
int schedule_coupled_IY_PE(config_t *config) { 
  return schedule_coupled_proactif(config, IY,"PE");
}   

/* schedule_coupled_IY_PR() */
int schedule_coupled_IY_PR(config_t *config) {
  return schedule_coupled_proactif(config, IY,"PR");
}     

/* schedule_coupled_IY_PY() */
int schedule_coupled_IY_PY(config_t *config) {
  return schedule_coupled_proactif(config, IY,"PY");
}

/* pick_IP_inc_up() */
static host_t *pick_inc_host_up(const char *alg, host_t **dont_pick, task_t *task)
{
  host_t *picked_host=NULL;
  double weight_picked, weight;
  int i;

  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = global.platform.hosts[i];

    /* If host is in the "don't pick" list, forget it */
    if (is_host_in_list(host, dont_pick))
      continue;

    /* If host isn't up, move on */
    if (STATE(host) != HOST_UP)
      continue;

   assign_task_to_host(task,host);

   if (!strcmp(alg,"IP")) {
     weight=coupled_success_proba();
   } else if (!strcmp(alg,"IE")) {
     weight=-coupled_exp_completion_time();
   } else if (!strcmp(alg,"IR")) {
     weight=coupled_success_proba()/coupled_exp_completion_time();
   }else if (!strcmp(alg,"IY")) {
     weight=coupled_success_proba()/(coupled_exp_completion_time()+(double)(global.current_date-global.application.date_of_current_iteration));
   }

    if((picked_host == NULL) ||(weight_picked<weight)){
      weight_picked=weight;
      picked_host=host;
    }
    unassign_task(task);
  }
  return picked_host;
}

/*communication_time_coupled()*/
static long communication_time_coupled(host_t *host)
{
  double weight_comm=0.0;
  int i;
  for (i=0; i < global.config->num_tasks; i++) {
    task_t *task = global.config->tasks[i];
    if ((task->activated) && (task->assigned_host == host))
    {
      weight_comm+= (task->time_to_data_download);
    }
  }
  if(host->data_download_time>weight_comm){weight_comm=0;}
  else{weight_comm-=host->data_download_time;}
  if(host->num_assigned_tasks>0){weight_comm+=host->time_to_prog_download;}
  return weight_comm;
}

static double P_nD(host_t *host,long t)
{
  double proba;
  proba=(1.0-(host->chain.matrix[0][2])-eigenval2(host))/(eigenval1(host)-eigenval2(host))*pow(eigenval1(host),t)+(1-(host->chain.matrix[0][2])-eigenval1(host))/(eigenval2(host)-eigenval1(host))*pow(eigenval2(host),t);
  return proba;
}

/*weight_on_host()*/
static double weight_on_host(host_t *host)
{
  double weight=0;
  int i;
  for (i=0; i < global.config->num_tasks; i++) {
    task_t *task = global.config->tasks[i];
    if ((task->activated) && (task->assigned_host == host))
    {
      weight+= (double)(host->w)*(task->remaining_work_fraction);
    }
  }
  return weight;
}

/*eigenval1()*/
static double eigenval1(host_t *host)
{
  double delta,x;
  //double matrix[3][3];
  //matrix=host->chain.matrix;
  delta=pow((host->chain.matrix[0][0]-host->chain.matrix[1][1]),2)+4.0*host->chain.matrix[0][1]*host->chain.matrix[1][0];
  x=(host->chain.matrix[0][0]+host->chain.matrix[1][1]+sqrt(delta))/2.0;
  return x;
}

/*eigenval2()*/
static double eigenval2(host_t *host)
{
  double delta,x;
  //double matrix[3][3];
  //matrix=host->chain.matrix;
  delta=pow((host->chain.matrix[0][0]-host->chain.matrix[1][1]),2)+4.0*host->chain.matrix[0][1]*host->chain.matrix[1][0];
  x=(host->chain.matrix[0][0]+host->chain.matrix[1][1]-sqrt(delta))/2.0;
  return x;
}

/*alpha1()*/
static double alpha1(host_t *host)
{
  double delta,x,sdelta;
  //double matrix[3][3];
  //matrix=host->chain.matrix;
  delta=pow(((host->chain).matrix[0][0]-(host->chain).matrix[1][1]),2)+4.0*(host->chain).matrix[0][1]*host->chain.matrix[1][0];
  sdelta=sqrt(delta);
  x=1/2.0+(host->chain.matrix[0][0]-host->chain.matrix[1][1])/(2.0*sqrt(delta));
  return x;
}

/*alpha2()*/
static double alpha2(host_t *host)
{
  double delta,x;
  //double matr[3][3];
  //matr=(host->chain).matrix;
  delta=pow((host->chain.matrix[0][0]-host->chain.matrix[1][1]),2)+4.0*host->chain.matrix[0][1]*host->chain.matrix[1][0];
  x=1/2.0-(host->chain.matrix[0][0]-host->chain.matrix[1][1])/(2.0*sqrt(delta));
  return x;
}

/*compute_expected_uS_coupled()*/
/*approximate the expected number of time slot with all processor UP before one
processor fails*/
static double compute_expected_uS_coupled(host_t **set_hosts,int set_host_size,double precision)
{
  double eus=0, current, prodt;
  double *e1t, *e2t;
  int i=0;
  double prod_eigenval_max=1.0;
  double pemt=1.0;
  if(set_host_size==0){return 0;}
  e1t=malloc(set_host_size*sizeof(double));
  e2t=malloc(set_host_size*sizeof(double));
  for(i=0;i<set_host_size;i++){
    prod_eigenval_max*=eigenval1(set_hosts[i]);
    e1t[i]=1.0;
    e2t[i]=1.0;
  }
  double bound=precision*(1.0-prod_eigenval_max);
  while(pemt>bound)
  {
    prodt=1.0;
    pemt*=prod_eigenval_max;
    for(i=0;i<set_host_size;i++){
      e1t[i]*=eigenval1(set_hosts[i]);
      e2t[i]*=eigenval2(set_hosts[i]);
      current=alpha1(set_hosts[i])*e1t[i]+alpha2(set_hosts[i])*e2t[i];
      prodt*=current;
    }
    eus+=prodt;
  }
  if(i>0){
    free(e2t);
    free(e1t);
  }
  return eus;
}

/*compute_expected_AS_coupled()*/
/*approximate the value A(S)*/
static double compute_expected_AS_coupled(host_t **set_hosts,int set_host_size,double precision)
{
  double as=0.0, current, prodt;
  double *e1t=NULL, *e2t=NULL;
  int i=0;
  double prod_eigenval_max=1.0;
  double pemt=1.0;
  if(set_host_size==0){return 0;}
  e1t=malloc(set_host_size*sizeof(double));
  e2t=malloc(set_host_size*sizeof(double));
  for(i=0;i<set_host_size;i++){
    prod_eigenval_max*=eigenval1(set_hosts[i]);
    e1t[i]=1.0;
    e2t[i]=1.0;
  }
  int t=0;
  while(pemt*((double)t/(1.0-prod_eigenval_max)+prod_eigenval_max/pow(1.0-prod_eigenval_max,2))>precision)
  {
    t++;
    prodt=1.0;
    pemt*=prod_eigenval_max;
    for(i=0;i<set_host_size;i++){
      e1t[i]*=eigenval1(set_hosts[i]);
      e2t[i]*=eigenval2(set_hosts[i]);
      current=alpha1(set_hosts[i])*e1t[i]+alpha2(set_hosts[i])*e2t[i];
      prodt*=current;
    }
    as+=t*prodt;
  }
  if(i>0){
    free(e2t);
    free(e1t);
  }
  return as;
}

/*compute_expected_Pfanny_coupled()*/
/*compute the probability of success of a computation on hosts set_hosts with a given precision*/
static double compute_expected_Pfanny_coupled(host_t **set_hosts, int set_host_size,double precision, int weight)
{
  if(set_host_size==0){return 0;}
  double us=compute_expected_uS_coupled(set_hosts,set_host_size,precision);
  double pplus=(us)/(1.0+us);
  return pow(pplus,weight-1);
}

/*compute_expected_Efanny_coupled*/
/*compute the expected computation time of a computation on hosts set_hosts with a given precision*/
static double compute_expected_Efanny_coupled(host_t **set_hosts, int set_host_size, double precision, int weight)
{
	if(set_host_size==0){return 0;}
  double us=compute_expected_uS_coupled(set_hosts,set_host_size,precision);
  double as=compute_expected_AS_coupled(set_hosts,set_host_size,precision);
  double ecs=as*(1.0-((us)/(1.0+us)))/(1.0+us);
  return (1.0+ecs*(weight-1.0))/(pow(us/(1.0+us),weight-1));
}

static double coupled_exp_completion_time()
{
  double expcost, expcomm, exp_num_up=0.0;
  long max_comm=0;
  double max_weight=0.0,weight,sum_comm=0.0,my_comm;
  int i,set_host_size=0;
  host_t **set_hosts=NULL;

  /*construct set_hosts,max_comm,sum_comm and max_weight*/
  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = global.platform.hosts[i];
    if (host->num_assigned_tasks>0)
    {
      my_comm=compute_expected_ct(host, communication_time_coupled(host));
      if (weight_on_host(host)>max_weight){max_weight=weight_on_host(host);}
      if (max_comm<my_comm){max_comm=my_comm;}
      sum_comm+=communication_time_coupled(host);
      set_hosts=realloc(set_hosts,(set_host_size+1)*sizeof(host_t*));
      set_hosts[set_host_size]=host;
      set_host_size++;
      exp_num_up+=(host->chain.matrix[1][0])/(host->chain.matrix[0][1]+host->chain.matrix[1][0]);
    }
  }
  if(exp_num_up>global.platform.max_num_comms){exp_num_up=global.platform.max_num_comms;}
  if(set_host_size<=global.platform.max_num_comms){exp_num_up=global.platform.max_num_comms;}
  sum_comm=sum_comm/exp_num_up;

  expcomm=max_comm;
  if (expcomm<sum_comm){expcomm=sum_comm;}
  weight=max_weight;
  expcost=compute_expected_Efanny_coupled(set_hosts,set_host_size,.01, weight)+expcomm;
  free(set_hosts);
  return expcost;
}

static double coupled_success_proba()
{
  double expcost=1.0, expcomm, sum_comm=0.0,my_comm;
  long max_comm=0;
  double max_weight=0.0,weight;
  int i,set_host_size=0,exp_num_up=0;
  host_t **set_hosts=NULL;

  /*construct set_hosts,max_comm,sum_comm and max_weight*/
  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = global.platform.hosts[i];
    if (host->num_assigned_tasks>0)
    {
      my_comm=compute_expected_ct(host, communication_time_coupled(host));
      if (weight_on_host(host)>max_weight){max_weight=weight_on_host(host);}
      if (max_comm<my_comm){max_comm=my_comm;}
      sum_comm+=communication_time_coupled(host);
      set_hosts=realloc(set_hosts,(set_host_size+1)*sizeof(host_t*));
      set_hosts[set_host_size]=host;
      set_host_size++;
      exp_num_up+=(host->chain.matrix[1][0]/(1.0-host->chain.matrix[1][3]))/(host->chain.matrix[0][1]/(1.0-host->chain.matrix[0][3])+host->chain.matrix[1][0]/(1.0-host->chain.matrix[1][3]));
    }
  }
  if(exp_num_up>global.platform.max_num_comms){exp_num_up=global.platform.max_num_comms;}
  if(set_host_size<=global.platform.max_num_comms){exp_num_up=global.platform.max_num_comms;}
  sum_comm=sum_comm/exp_num_up;

  expcomm=max_comm;
  if (expcomm<sum_comm){expcomm=sum_comm;}
  weight=max_weight;
  for (i=0; i < set_host_size; i++) {
    host_t *host = set_hosts[i];
    if(is_host_in_list(host,set_hosts))
    {
      my_comm=communication_time_coupled(host);
      expcost*=pow(host->Pfanny,my_comm);
      expcost*=P_nD(host,max_comm-compute_expected_ct(host, my_comm));
    }
  }
  expcost*=compute_expected_Pfanny_coupled(set_hosts,set_host_size,.01, weight);
  free(set_hosts);
  return expcost;
}
