#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "simulator.h"

static void simulation_init(void);
static int simulation_finished(void);
static int can_coupled_progress(void);
static void advance_simulation_clock(void);
static int update_pending_communications(void);
static int update_pending_computations(void);
static int update_pending_computations_coupled(void);
static int update_pending_computations_independent(void);
static int is_iteration_complete(void);
static void start_new_iteration(void);
static int is_coupled_compromised(void);
static void unassign_tasks_from_down_hosts(void);
static void unassign_all_non_started_tasks(void);
static void hosts_init(void);
static int cancel_clones(void);
static int num_clones(task_t *task);
static void print_simulation_state(void);
static int all_tasks_at_same_fraction(void);
static double min_fraction(void);

static int compare_picked_hosts(const void *, const void*);

static double compute_current_available_compute_power(void);
static double compute_current_used_compute_power(void);

/* schedule_init() */
static void simulation_init(void) {

  /* Initialize application */
  global.application.num_completed_iterations = 0;
  global.application.date_of_current_iteration = 0;

  /* Initialize hosts  */
  /* All hosts start at the beginning of the trace */
  int i;
  for (i=0; i < global.platform.num_hosts; i++) {
    global.platform.hosts[i]->availability.current_entry = 
        global.platform.hosts[i]->availability.start_entry;
  }

  /* Initialize counters */
  global.available_compute_power = 0.0;
  global.used_compute_power      = 0.0;
  global.wasted_compute_power    = 0.0;
  global.comm_timesteps          = 0;
  global.comm_wasted_timesteps   = 0;

  return;
}

/* run_simulation() */
int run_simulation(algorithm_t alg)
{
  /* Initialize global counters */
  simulation_init();

  /* Initialize the tprogs */
  hosts_init();

  /* Initialize configuration */
  if (global.config)
    free_config(global.config);
  global.config = new_config();

  /* Set the simulation time to zero */
  global.current_date = 0;

  /* Go through the time steps */
  while (1) {

    /* Unassign all tasks that have not begun, so that
       they can be reconsidered anew by the scheduling algorithms */
    if (global.application.mode == MODE_INDEPENDENT)
    {unassign_all_non_started_tasks();}

    /* Invoke the scheduling algorithm */
    /* And pass it the config          */
    if ((schedule_functions[alg])(global.config))
      return -1;
  
    /* Print host and task states for verbose info/debug */
    print_simulation_state();

    /* Update some counter */
    global.available_compute_power += 
                  compute_current_available_compute_power();
    global.used_compute_power += 
                  compute_current_used_compute_power();

    /* Update pending computations              
     * Doing the computations before the communications to avoid
     * weird "dependency" issues
     */
    if (update_pending_computations())
      return -1;

    /* Deal with clones */
    if (cancel_clones())
      return -1;

    /* Update pending communications */
    if (update_pending_communications())
      return -1;

    /* Advance iteration count */
    if (is_iteration_complete()) {
      /* Update the iteration count */
      PRINT_VERBOSE0("*Iteration Completed*\n");
      global.application.num_completed_iterations++;
    }

    /* Check whether we should end here */
    if (simulation_finished()){
      free_config(global.config);
      break;
    }
    
    /* Re-check if the iteration is finished () */
    /* Sort of ugly, but who cares */
    if (is_iteration_complete()) { 
      start_new_iteration(); 
    }

    /* Advance clock */
    advance_simulation_clock();

    /* Unassign all tasks that were on down hosts */
    unassign_tasks_from_down_hosts();

    if (global.application.mode == MODE_COUPLED)
    {
    all_tasks_at_same_fraction();
    }


  }
  return 0;
}

/* simulation_finished() */
static int simulation_finished(void)
{
  if (global.application.max_num_iterations != -1) {
    if (global.application.num_completed_iterations >=
        global.application.max_num_iterations) {
      return 1;
    }
  } 
  if (global.application.timelimit != -1) {
    if (global.current_date >= 
        global.application.timelimit) {
      return 1;
    }
  }
  return 0;
}

/* can_coupled_progress() */
static int can_coupled_progress(void)
{
  int i;

  for (i=0; i < global.config->num_tasks; i++) {
    task_t *task = global.config->tasks[i];
    host_t *host = task->assigned_host;
    if (!task->activated)
      continue;
    if ((!host) ||
        (STATE(host) != HOST_UP) ||
        (host->time_to_prog_download > 0) || 
        (task->time_to_data_download > 0)) {
      return FALSE;
    }
  }
  return TRUE;
}

/* advance_simulation_clock() 
 *  Implements wrap-around of availability traces
 */
static void advance_simulation_clock(void)
{
  int i;

  /* update the date */
  (global.current_date)++;

  /* update the current available entry for each host, possibly
   * implementing a wrap-around.
   */
  for (i=0; i < global.platform.num_hosts; i++) {
    availability_trace_t *trace = &(global.platform.hosts[i]->availability);

    if (trace->current_entry == trace->num_entries-2) {
      /* Need to do a wrap-around */
      /* Note that the last entry is bogus because we don't know
         how long it lasts, so we wrap around back to the first entry,
         which is a small error because this transition wasn't generated
         by the markov chain or by the trace. This could be fixed by not
         storing dates, but instead durations, in availability traces */
      trace->current_entry = 0;
      trace->time_offset += trace->dates[trace->num_entries-1];
    } else if (trace->dates[trace->current_entry+1] + trace->time_offset  >
          global.current_date) {
      /* Do nothing */
    } else {
      /* Simple advance */
      (trace->current_entry)++;
    }
  }
  return;
}

/* update_pending_communications() 
 *  This routine will allow up to max_num_comms communications
 *  to make progress at the same time, chosen in (arbitrary) order
 *  of hosts and, for a given host, tasks.  NEVER any simultaneous
 *  communications to a given host.
 **/
static int update_pending_communications(void)
{
  int num_comms;
  int i,j;

  /****************************************************************/
  /* Deal with down hosts and reinitialize their program download */
  /****************************************************************/
  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = global.platform.hosts[i];
    /* If host isn't down, forget it */
    if (STATE(host) != HOST_DOWN) 
      continue;

    /* Account for waste in case download is interrupted */
    if ((host->time_to_prog_download < global.application.Tprog) &&
        (host->time_to_prog_download > 0)) {
      global.comm_wasted_timesteps += 
               global.application.Tprog - host->time_to_prog_download;
    }

    /* No matter what, the program needs to be downloaded again */
    host->time_to_prog_download = global.application.Tprog;
    host->data_download_time=0;

    /* Note that we do not deal with the Tdata thing here, because
     * it's dealt with in unassign_tasks_from_down_hosts  (tasks that
     * become unscheduled see their Tdata reset)
     */
  }

  /*****************************/
  /* Deal with reclaimed hosts */
  /*****************************/
  if (global.platform.nonresumabledownloads) {
    for (i=0; i < global.platform.num_hosts; i++) {
      host_t *host = global.platform.hosts[i];
      /* If host isn't reclaimed, forget it */
      if (STATE(host) != HOST_RECLAIMED)
        continue;
  
      /* If the program was being downloaded, the download
         must be resumed from scratch */
      /* account for waste */
      if ((host->time_to_prog_download > 0) &&
          (host->time_to_prog_download < global.application.Tprog)) {
        global.comm_wasted_timesteps += host->time_to_prog_download;
      }
      if (host->time_to_prog_download > 0) {
        host->time_to_prog_download = global.application.Tprog;
      }
  
      /* Pending data communications are gone as well unless
         they are for completed tasks */
      for (j=0; j < global.config->num_tasks; j++) {
        task_t *task = global.config->tasks[i];

        /* If the task isn't active, forget it */
        if (!task->activated)
          break;

        /* If the task isn't scheduled on the host, forget it */
        if (task->assigned_host != host)
          continue;

        /* If the task hasn't completed, then something may need to be done */
        if (!task->done) {
          /* If the data hadn't been completely downloaded, redownload it */
          if (task->time_to_data_download > 0) {
            if (task->time_to_data_download < global.application.Tdata) { 
              global.comm_wasted_timesteps += task->time_to_data_download;
            }
            task->time_to_data_download = global.application.Tdata;
          } 
        } 
      }
    }
  } else {
    /* NOTHING TO DO */
  }

  /**********************/
  /* Deal with up hosts */
  /**********************/
  num_comms = 0;

  /* Initialize the has_communicated_this_step counter */
  for (i=0; i < global.platform.num_hosts; i++) {
    global.platform.hosts[i]->has_communicated_this_step = FALSE;
  }

  /* Sort the hosts according to the picking order */
  host_t **sorted_hosts;
  sorted_hosts = (host_t **)calloc(global.platform.num_hosts, sizeof(host_t*));
  for (i=0; i < global.platform.num_hosts; i++) {
    sorted_hosts[i] = global.platform.hosts[i];
  }

  qsort(sorted_hosts, global.platform.num_hosts, sizeof(host_t*),
        compare_picked_hosts);

  
  /* Deal with currently ongoing communications, one per host, taken
   * in the picking order */
  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = sorted_hosts[i];

    // If we have reached the total number of comms, we're done
    if (num_comms >= global.platform.max_num_comms)
    {
      free(sorted_hosts);
      return 0;
    }

    // If the host isn't up, forget it
    if (STATE(host) != HOST_UP) 
      continue;

    // If the host isn't used in the schedule, forget it
    if (num_tasks_assigned_to_host(host) == 0) 
      continue;

    // If the host doesn't have the program, then it makes progress
    if (host->time_to_prog_download > 0) {
      host->time_to_prog_download--;
      host->has_communicated_this_step = TRUE;
      global.comm_timesteps++;
      num_comms++;
      continue; // Only one comm per host
    } 

    // Pick a task on whose comm to make progress
    task_t *picked_task = NULL;
    for (j=0; j < global.config->num_tasks; j++) {
      task_t *task = global.config->tasks[j];

      // If the task isn't activated, forget it
      if (!task->activated)
        continue;
   
      // If the task isn't scheduled on the host, forget it 
      if (task->assigned_host != host) 
        continue;

      if ( ((task->time_to_data_download > 0) &&
            (task->time_to_data_download < global.application.Tdata))
           &&
          ((picked_task == NULL) || (num_clones(task) < num_clones(picked_task))) ) {
        picked_task = task;
      }
    }
    if (picked_task) {  // If found a task to make progress on
      picked_task->time_to_data_download--;
      host->has_communicated_this_step = TRUE;
      host->data_download_time++;
      global.comm_timesteps++;
      num_comms++;
    }
  }

  /* Deal with not-yet-started communications, one per host, taken
   * in the picking order */
  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = sorted_hosts[i];

    // If we have reached the total number of comms, we're done
    if (num_comms >= global.platform.max_num_comms)
    {
      free(sorted_hosts);
      return 0;
    }

    // If the host isn't up, forget it
    if (STATE(host) != HOST_UP) 
      continue;

    // If the host doesn't have the program, forget it 
    if (host->time_to_prog_download > 0) 
      continue;
    
    // If the host has communicated this step, forget it
    if (host->has_communicated_this_step)
      continue;

    // Can we start making progress on a comm that hasn't started yet?
    // We can if there is NO currently assigned task that has:
    //   - time_to_data_download = 0
    //   - remaining_work_fraction = 1.0
    // This is because we don't start downloading "too" ahead of time.
    for (j=0; j < global.config->num_tasks; j++) {
      task_t *task = global.config->tasks[j];

      // If the task isn't activated, forget it
      if (!task->activated)
        continue;
   
      // If the task isn't scheduled on the host, forget it 
      if (task->assigned_host != host) 
        continue;
      
      if (global.application.mode == MODE_INDEPENDENT) {
        if ((task->time_to_data_download == 0) && 
            (task->remaining_work_fraction >= 1.0)) {
          break;
        }
      }
    }
    // If there is already a task with (0,1.0), done for this host
    // Because we don't start downloading task n+1 because task n
    // has begun computation
    if (j < global.config->num_tasks)
      continue;

    /* Pick a task that hasn't started downloading yet and on which
       we can make progress */
    task_t *picked_task = NULL;
    for (j=0; j < global.config->num_tasks; j++) {
      task_t *task = global.config->tasks[j];

      // If the task isn't activated, forget it
      if (!task->activated)
        continue;
   
      // If the task isn't scheduled on the host, forget it 
      if (task->assigned_host != host) 
        continue;

      if ( (task->time_to_data_download > 0) && 
           ((picked_task == NULL) || 
            (num_clones(task) < num_clones(picked_task)))) {
        picked_task = task;
      }
    }
    if (picked_task) { // We found one 
      picked_task->time_to_data_download--;
      host->has_communicated_this_step = TRUE;
      host->data_download_time++;
      global.comm_timesteps++;
      num_comms++;
    }
  }

  // If we have reached the total number of comms, we're done
  if (num_comms >= global.platform.max_num_comms) 
  {
    free(sorted_hosts);
    return 0;
  }

  /* Make progress on hosts that don't currently have a task but that 
   * are in the middle of downloading the program... doesn't hurt 
   */
  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = sorted_hosts[i];

    // If the host isn't up, forget it
    if (STATE(host) != HOST_UP) 
      continue;

    if (host->has_communicated_this_step)
      continue;

    // If the host doesn't have the program but is in the process of
    // downloading it, then do it
    if ((host->time_to_prog_download > 0) && 
        (host->time_to_prog_download < global.application.Tprog)){
      host->time_to_prog_download--;
      host->has_communicated_this_step = TRUE;
      global.comm_timesteps++;
      num_comms++;
      continue; // Only one comm per host
    }
  }

  // If we have reached the total number of comms, we're done
  if (num_comms >= global.platform.max_num_comms) 
  {
      free(sorted_hosts);
       return 0;
  }

  /* Make progress on hosts that don't currently have a task, and haven't
   * started downloading the program, but could... doesn't hurt 
   */
  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = sorted_hosts[i];

    // If the host isn't up, forget it
    if (STATE(host) != HOST_UP) 
      continue;

    if (host->has_communicated_this_step)
      continue;

    // If the host doesn't have the program, then start downloading it
    if (host->time_to_prog_download == global.application.Tprog){
      host->time_to_prog_download--;
      host->has_communicated_this_step = TRUE;
      global.comm_timesteps++;
      num_comms++;
      continue; // Only one comm per host
    }
  }
  free(sorted_hosts);
  return 0;
}

/* is_iteration_complete() */
static int is_iteration_complete(void)
{
  int i,j;

  for (i=0; i < global.application.num_tasks; i++) {
    int done = FALSE;
    for (j=0; j < global.platform.num_hosts; j++) {
      task_t *task = global.config->tasks[i*global.platform.num_hosts + j];
      if ((task->activated) && (task->done)) {
        done = TRUE;
        break;
      }
    }
    if (!done)
      return FALSE;
  }
  return TRUE;
}

/* start_new_iteration() */
static void start_new_iteration(void)
{
  int i,j;

  /* Simply reset all tasks */
  for (i=0; i < global.application.num_tasks; i++) {
    task_t *task = global.config->tasks[i*global.platform.num_hosts];
    if (task->assigned_host) {
      unassign_task(task);
    }
    task->remaining_work_fraction = 1.0;
    task->time_to_data_download = global.application.Tdata;
    task->activated = TRUE;
    task->done = FALSE;

    for (j=1; j < global.platform.num_hosts; j++) {
      task_t *clone_task = global.config->tasks[i*global.platform.num_hosts + j];
      if (clone_task->assigned_host) {
        unassign_task(clone_task);
      }
      clone_task->remaining_work_fraction = 1.0;
      clone_task->time_to_data_download = global.application.Tdata;
      clone_task->activated = FALSE;
      clone_task->done = FALSE;
    }
    global.application.date_of_current_iteration=global.current_date+1;
  }
  return;
} 

/* update_pending_computations */
static int update_pending_computations(void)
{
  if (global.application.mode == MODE_COUPLED) {
    return update_pending_computations_coupled();
  }  else if (global.application.mode == MODE_INDEPENDENT) {
    return update_pending_computations_independent();
  } 
  fprintf(stderr,"Error: Unknown application mode!\n");
  return -1;
}

/* update_pending_computations_coupled() */
static int update_pending_computations_coupled(void)
{
  int i;

  /* If a task is unscheduled, then everybody will need to be 
     restarted from scracth */
  if (is_coupled_compromised()) {
    for (i=0; i < global.application.num_tasks; i++) {
      task_t *task = global.config->tasks[i * global.platform.num_hosts]; // Skip clones
      global.wasted_compute_power += (1.0 - task->remaining_work_fraction);
      task->remaining_work_fraction = 1.0;
    }
    return 0;
  }

  /* If no progress is possible, forget it */
  if (!can_coupled_progress())
   return 0;

  /* At this point, everybody makes progress */
  for (i=0; i < global.application.num_tasks; i++) {
    task_t *task = global.config->tasks[i * global.platform.num_hosts]; // Skip clones
    //host_t *host = task->assigned_host;

    // XXX: should be the speed of the slowest processor!!!!
    task->remaining_work_fraction -= min_fraction();
    global.used_compute_power += min_fraction();
    if (task->remaining_work_fraction < EPSILON) {
      task->remaining_work_fraction = 0;
      task->done = TRUE;
      unassign_task(task);
    }
  }

  return 0;
}

/* update_pending_computations_independent() */
static int update_pending_computations_independent(void)
{
  int i,j;

  /* Deal with up hosts */
  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = global.platform.hosts[i];

    /* If the host can't compute, move on */
    if ((STATE(host) != HOST_UP) || 
        (host->time_to_prog_download > 0))
      continue;

    /* Otherwise, decrease the currently being computed tasks, if any, or
       a not-yet-begun task yet */
    task_t *picked_task = NULL;
    for (j=0; j < global.config->num_tasks; j++) {
      task_t *task = global.config->tasks[j];

      /* If the task isn't activated, forget it */
      if (!task->activated)
        continue;

      /* If the task isn't scheduled on the host, forget it */
      if (task->assigned_host != host)
        continue;

      /* If the task's data hasn't been downloaded, forget it */
      if (task->time_to_data_download > 0) 
        continue;

      /* If the task is done, forget it */
      if (task->done) 
        continue;

      if ((picked_task == NULL) ||
          (task->remaining_work_fraction < picked_task->remaining_work_fraction) || 
          ((fabs(task->remaining_work_fraction - picked_task->remaining_work_fraction) < EPSILON)  && 
           (num_clones(task) < num_clones(picked_task)))) {
        picked_task = task; 
      }
    }

    /* Making progress on task_picked */
    if (picked_task) {
//      fprintf(stderr,"Decreasing computation of task %d on host %d (Tdata=%d)\n",picked_task->id,host->id,picked_task->time_to_data_download);
      picked_task->remaining_work_fraction -=  (1.0 / (double)(host->w));
      global.used_compute_power +=  (1.0 / (double)(host->w));
      if (picked_task->remaining_work_fraction < EPSILON) {
        picked_task->remaining_work_fraction = 0;
        picked_task->done = TRUE;
        unassign_task(picked_task);
      }
    } else {
//      fprintf(stderr,"No task can make progress on host %d\n",host->id);
    }
  }

  return 0;
}


/* unassign_task() */
void unassign_task(task_t *task)
{
  host_t *host = task->assigned_host;

  if (host == NULL) {
    fprintf(stderr,"Warning: unassigning unassigned task (error #1)\n");
    return;
  }

  /* Deal with the data-download time */
  if ((task->time_to_data_download > 0) && 
      (task->time_to_data_download < global.application.Tdata)) {
    global.comm_wasted_timesteps += 
             global.application.Tdata - task->time_to_data_download;
  }

  /* Deal with the compute fraction */
  if ((task->remaining_work_fraction > EPSILON) &&
      (task->remaining_work_fraction < 1.0)) {
    global.wasted_compute_power += 1.0 - task->remaining_work_fraction;
  }

  /* If the task was unassigned but not done, then reset it */
  if (!task->done) {
    task->time_to_data_download = global.application.Tdata;
    task->remaining_work_fraction = 1.0;
  }

  task->assigned_host = NULL;
  host->num_assigned_tasks--;
  return;
}

/* assign_task_to_host() */
void assign_task_to_host(task_t *task, host_t *host)
{

  // Sanity check
  if (STATE(host) != HOST_UP) {
    fprintf(stderr,"Can only assign a task to a UP host\n");
    exit(1);
  }

  // Update the task's data
  task->time_to_data_download = global.application.Tdata;
  task->assigned_host = host;
  task->remaining_work_fraction = 1.0; // The task start from scratch

  host->num_assigned_tasks++;

  return;
}

/* is_coupled_compromised(void) */
static int is_coupled_compromised(void)
{
  int i;
  for (i=0; i < global.application.num_tasks; i++) {//config->application
    task_t *task = global.config->tasks[i * global.platform.num_hosts ];
    host_t *host;
    host = task->assigned_host;
    if (!host)
      return TRUE;
  }
  return FALSE;
}

/* compute_current_available_compute_power() */
static double compute_current_available_compute_power(void)
{
  int i;
  double power = 0.0;

  for (i=0; i < global.platform.num_hosts; i++) {
    if (STATE(global.platform.hosts[i]) == HOST_UP) {
      power += 1.0 / (global.platform.hosts[i]->w); 
    }
  }
  return power;
}

/* compute_current_used_compute_power() */
static double compute_current_used_compute_power(void)
{
  int i;
  double power = 0.0;

  for (i=0; i < global.platform.num_hosts; i++) {
    if ((STATE(global.platform.hosts[i]) == HOST_UP) &&
        (num_tasks_assigned_to_host(global.platform.hosts[i]) != 0)) {
      power += 1.0 / (global.platform.hosts[i]->w); 
    }
  }
  return power;

}

/* unassign_tasks_from_down_hosts() */
static void unassign_tasks_from_down_hosts(void)
{
  int i,removed_tasks=0;

  for (i=0; i < global.config->num_tasks; i++) {
    task_t *task = global.config->tasks[i];
    /* If the task isn't activated, forget it */
    if (!task->activated)
      continue;
    /* If the task is not assigned, forget it */
    if (!task->assigned_host)
      continue;
    if (STATE(task->assigned_host) == HOST_DOWN) {
      unassign_task(task);
      removed_tasks++;
    }
  }
  if (global.application.mode == MODE_COUPLED){
    if (removed_tasks>0)
    {
      for (i=0; i < global.config->num_tasks; i++) {
        task_t *task = global.config->tasks[i];
        if (task->assigned_host != NULL) {
    	  unassign_task(task);
    	}
      }
    }
  }
}

/* print_simulation_state() */
static void print_simulation_state(void)
{
  int i,j;

  char charstate[3]="URD";
  PRINT_VERBOSE1("Time %ld:\n",global.current_date);
  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = global.platform.hosts[i];
    availability_trace_t trace = host->availability; 
    PRINT_VERBOSE4("  host #%d: %c, ttprog=%d, %d tasks\n",
            i,
            charstate[(int)(trace.states[trace.current_entry])],
            host->time_to_prog_download,
            num_tasks_assigned_to_host(host));
    for (j=0; j < global.config->num_tasks; j++) {
      task_t *task = global.config->tasks[j];
      /* If the task isn't activated, forget it */
      if (!task->activated)
        continue;
      /* If the task isn't scheduled on the host, forget it */
      if (task->assigned_host != host)
        continue;
      if (task->clone_of) {
        PRINT_VERBOSE3("    task %d (clone): ttdata=%d  frac=%.2f\n",
                       task->id,
                       task->time_to_data_download,
                       task->remaining_work_fraction);
      } else {
        PRINT_VERBOSE3("    task %d (orig.): ttdata=%d  frac=%.2f\n",
                       task->id,
                       task->time_to_data_download,
                       task->remaining_work_fraction);
      }
    }
  }
}

/* free_config() */
void free_config(config_t *config)
{
  int j;
 
  for (j=0; j < global.config->num_tasks; j++)
    free(config->tasks[j]);
  free(config->tasks);

  free(config);
  return;
}

/* new_config */
config_t *new_config(void)
{
  int count,i,j;
  config_t *config;

  config = (config_t *)calloc(1,sizeof(config_t));

  /* Each task can have 1 original copy and n-1 clones, 
     each running on one of the n hosts */
  config->num_tasks = global.application.num_tasks * global.platform.num_hosts;

  config->tasks = (task_t **)calloc(config->num_tasks, sizeof(task_t*));

  /* Initialize the main tasks and their clones */
  count = 0;
  for (i=0; i < global.application.num_tasks; i++) {
    int original_task_index = i * global.platform.num_hosts;
    config->tasks[original_task_index] = (task_t *)calloc(1,sizeof(task_t));
    config->tasks[original_task_index]->id = i;
    config->tasks[original_task_index]->clone_of = NULL;
    config->tasks[original_task_index]->activated = TRUE;
    config->tasks[original_task_index]->done = FALSE;
    config->tasks[original_task_index]->assigned_host = NULL;
    config->tasks[original_task_index]->time_to_data_download = global.application.Tdata;
    config->tasks[original_task_index]->remaining_work_fraction = 1.0;
    for (j=1; j < global.platform.num_hosts; j++) {
      int clone_task_index = original_task_index + j;
      config->tasks[clone_task_index] = (task_t *)calloc(1,sizeof(task_t));
      config->tasks[clone_task_index]->id = i;
      config->tasks[clone_task_index]->clone_of = config->tasks[original_task_index];
      config->tasks[clone_task_index]->activated = FALSE;
      config->tasks[clone_task_index]->done = FALSE;
      config->tasks[clone_task_index]->assigned_host = NULL;
      config->tasks[clone_task_index]->time_to_data_download = global.application.Tdata;
      config->tasks[clone_task_index]->remaining_work_fraction = 1.0;
    }
  }

  return config;
}

/*min_fraction*/
static double min_fraction()
{
  int i;
  double count=1;
  for(i=0;i<global.platform.num_hosts;i++)
  { if(num_tasks_assigned_to_host(global.platform.hosts[i])>0){
	if(1/((double)num_tasks_assigned_to_host(global.platform.hosts[i])*(double)global.platform.hosts[i]->w)<count)
	{count=1/((double)num_tasks_assigned_to_host(global.platform.hosts[i])*(double)global.platform.hosts[i]->w);}
  }}
  return count;
}


/* num_tasks_asssigned_to_host() 
 *  counts the task being executed and all tasks after it
 */
int num_tasks_assigned_to_host(host_t *host)
{
  return host->num_assigned_tasks;
}

/* num_tasks_not_yet_started_on_host(host_t *host) */
int num_tasks_not_yet_started_on_host(host_t *host)
{
  int i;
  int num=0;

  for (i=0; i < global.config->num_tasks; i++) {
    task_t *task =global.config->tasks[i];

    /* If the task is not activated, fine */
    if (!task->activated)
      continue;

    /* If the task is not assigned, fine */
    if (task->assigned_host != host)
      continue;

    /* If the task hasn't started, count it */
    if (task->remaining_work_fraction == 1.0)
      num++;
  }
  return num;
}

/* num_tasks_not_yet_completed_on_host() */
int num_tasks_not_yet_completed_on_host(host_t *host)
{
  int i;
  int num=0;

  for (i=0; i < global.config->num_tasks; i++) {
    task_t *task =global.config->tasks[i];

    /* If the task is not activated, fine */
    if (!task->activated)
      continue;

    /* If the task is not assigned, fine */
    if (task->assigned_host != host)
      continue;

    /* If the task is done, count it */
    if (task->remaining_work_fraction > EPSILON)
      num++;
  }
  return num;
}

/* unassign_all_non_started_tasks() */
static void unassign_all_non_started_tasks(void)
{
  int i;

  for (i=0; i < global.config->num_tasks; i++) {
    task_t *task =global.config->tasks[i];

    /* If the task is not activated, fine */
    if (!task->activated)
      continue;

    /* If the task is done, fine */
    if (task->done)
      continue;

    /* If the task is not assigned, fine */
    if (!task->assigned_host)
      continue;
  
    /* If the task has begun communication, fine */
    if (task->time_to_data_download < global.application.Tdata)
      continue;
  
    /* If the task has begun computation, fine (useless, but just in case) */
    if (task->remaining_work_fraction < 1.0)
      continue;

    /* Unassign the task */
    unassign_task(task);
  }

  return;
}

static int compare_picked_hosts(const void *h1, const void *h2)
{
  host_t **host1 = (host_t **)h1;
  host_t **host2 = (host_t **)h2;

  return ((*host1)->picking_order - (*host2)->picking_order);
}

static void hosts_init(void) {
  int i;

  for (i=0; i < global.platform.num_hosts; i++) {
    host_t *host = global.platform.hosts[i];
    host->time_to_prog_download = global.application.Tprog;
    host->num_assigned_tasks = 0;
    host->data_download_time=0;
  }
  return;
}

/* cancel_clones() */
static int cancel_clones()
{
  int i,j;

  if (global.max_num_clones == 0)
    return 0;

  for (i=0; i < global.application.num_tasks; i++) {
    int done = FALSE;
    for (j=0; j < global.platform.num_hosts; j++) {
      int task_index = i * global.platform.num_hosts + j;  
      task_t *task = global.config->tasks[task_index];
      done = done || task->done;
    }
    if (done) {
      /* Unassign all corresponding tasks */
      for (j=0; j < global.platform.num_hosts; j++) {
        task_t *task = global.config->tasks[i * global.platform.num_hosts + j];
        if (task->assigned_host) {
          unassign_task(task);
        }
        task->remaining_work_fraction = 0.0;
        task->done = TRUE;
        if (j > 0) {
          task->activated = FALSE;
        }
      }
    }
  }
  return 0;
}

/* number of clones */
static int num_clones(task_t *task)
{
  int j;
  int num = 0;

  if (global.max_num_clones == 0)
    return 1;

  for (j=0; j < global.platform.num_hosts; j++) {
    task_t *clone = global.config->tasks[task->id * global.platform.num_hosts + j];
    if ((clone->activated) && (!clone->done))
      num++;
  }
  return num;
}


static int all_tasks_at_same_fraction()
{
	int i;
	double frac_max=0;
	for (i=0; i < global.application.num_tasks; i++) {
		task_t *task = global.config->tasks[i*global.platform.num_hosts];
		if (frac_max < task->remaining_work_fraction)
		{frac_max=task->remaining_work_fraction;}
	}
	for (i=0; i < global.application.num_tasks; i++) {
		task_t *task = global.config->tasks[i*global.platform.num_hosts];
		task->remaining_work_fraction=frac_max;
	}
	return 0;
}

