#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "oshseed.h"
#define ASENV_SHORT_NAMES
#include "envseed.h"
#define SC ASOSH_SWITCH_CHARACTER

#include "availability.h"

#define EPSILON 0.0000001

struct {
  int verbose;
  distribution_t distribution;
  double distribution_shape;
  int seed;
  int length;
  markov_chain_t chain;
} global;

/*
 * Prototypes 
 */
static int parseOptions(const char * const *argv);
static void printUsage(void);
static int BadSwitchHandler(ParseErrors error,
                 const char *badWord, unsigned badSwitchOffset,
                 unsigned badSwitchLen, const char *badValue);

static int BadSwitchSeen=0;

/*
 * Command-line arguments
 */
int main(int argc, char **argv) 
{
  int option_status;
  availability_trace_t trace;

  /* parse command-line arguments */
  if ((option_status = parseOptions((const char * const *)argv)) < 0) {
    if (option_status == -1)
      printUsage();
    exit(1);
  }

  /* Compute the steady-state distribution */
  computePiDistribution(&(global.chain));

  printChain(&global.chain);


  /* Build trace */
  if (buildTraceFromChain(global.seed,global.length, 
                          global.distribution, global.distribution_shape, 
                          &global.chain,&trace)) {
    exit(1);
  }

  /* Print trace */
  printTrace(&trace);

  /* Print additional information */
  if (global.verbose) {
    fprintf(stderr,"Steady State Distribution: pu=%lf pr=%lf pd=%lf\n",
	global.chain.pi[0], global.chain.pi[1], global.chain.pi[2]);
  }

  /* Sanity Check */
  {
    fprintf(stderr,"SANITY CHECK\n");
    markov_chain_t chain;
    buildChainFromTrace(&chain,&trace);
    printChain(&chain);
  }
  exit(0);
}

/* printUsage() */
static void printUsage(void)
{

  fprintf(stderr,"Usage: tracegen [flags]\n"
                 "  [-h]                					prints usage\n"
                 "  [-v]                					verbose mode\n"
                 "  [-seed <seed>]      					(default: seed=0)\n"
		 "  -markov <uu>#<ur>#<ud>#<ru>#<rr>#<rd>#<du>#<dr>#<dd>	Markov transitions\n"
		 "  [-pareto a]      Generate a Pareto trace based on the Markov sojourn times instead\n"
		 "  [-weibull a]      Generate a Weibull trace based on the Markov sojourn times instead\n"
		 "                   (using a Pareto distribution with shape parameter a)\n"
 		 "  [-length <length>]					trace length (default=100)\n"
          );
  return;
}

/*
 * A switch error handler.  Passed the error, the element of argv
 * that contains the bad switch, the offset into that element of the
 * switch name, the length of the switch name, and the bad value,
 * if appropriate.  
 */
static int BadSwitchHandler(ParseErrors error,
                 const char *badWord, unsigned badSwitchOffset,
                 unsigned badSwitchLen, const char *badValue) {

  int intLen = (int)badSwitchLen;

  if(error == ARGUMENTS_NOT_ALLOWED)
    fprintf(stderr, "This program takes no program arguments\n");
  else if(error == INVALID_VALUE)
    fprintf(stderr, "Error: Value '%s' is invalid for switch '%*.*s'\n",
            badValue, intLen, intLen, badWord + badSwitchOffset);
  else if(error == MISSING_VALUE)
    fprintf(stderr, "Error: Switch '%*.*s' requires a value\n",
            intLen, intLen, badWord + badSwitchOffset);
  else if(error == UNKNOWN_SWITCH)
    fprintf(stderr, "Error: '%*.*s' is not a known switch\n",
            intLen, intLen, badWord + badSwitchOffset);

  BadSwitchSeen = 1;
  return 1; /* Allow argv parsing to continue. */
}

/* Parse the options and set default values 
 *
 * returns -1 when the usage should be printed
 * returns -2 when the usage should not be printed
 * returns 0 when Ok
 */

static int parseOptions(const char * const *argv)
{
  /* Declaration of all possible options */
  const char *validSwitches =
    SC "h void             \n"
    SC "v void             \n"
    SC "seed int           \n"
    SC "length int         \n"
    SC "pareto double      \n"
    SC "weibull double      \n"
    SC "markov string      \n";

  ParsedArgv pa;

  /* Check for errors */
  if ((pa = ParseArgv(argv, validSwitches,&BadSwitchHandler)) == NULL ||
      BadSwitchSeen) {
    fprintf(stderr, "\n%s\n", validSwitches);
    return 1;
  }

  /* -h switch */
  if (SwitchPresent(pa,"-h"))
    return -1;

  /* -v switch */
  if (SwitchPresent(pa,"-v")) {
    global.verbose = 1;
  } else {
    global.verbose = 0;
  }

  /* -seed switch */
  if (SwitchPresent(pa,"-seed")) {
    const char *seed_string=SwitchValueN(pa,"-seed",1,"0");
    if (*SwitchValueN(pa,"-seed",2,""))
      fprintf(stderr,"Warning: multiple -seed flags, ignoring all but the first one\n");
    sscanf(seed_string,"%d",&(global.seed));
  }  else {
    fprintf(stderr,"Using default seed: 0\n");
    global.seed = 0;
  }

  /* -markov switch */
  if (SwitchPresent(pa,"-markov")) {
    const char *markov_string;
    if (*SwitchValueN(pa,"-markov",2,"")) 
      fprintf(stderr,"Warning: multiple -markov flags, ignoring all but the first one\n");

    markov_string = SwitchValueN(pa,"-markov",1,"");
    if (sscanf(markov_string,"%lf#%lf#%lf#%lf#%lf#%lf#%lf#%lf#%lf",
		&(global.chain.matrix[0][0]),&(global.chain.matrix[0][1]),&(global.chain.matrix[0][2]), 
		&(global.chain.matrix[1][0]),&(global.chain.matrix[1][1]),&(global.chain.matrix[1][2]), 
		&(global.chain.matrix[2][0]),&(global.chain.matrix[2][1]),&(global.chain.matrix[2][2])) != 9) {
      fprintf(stderr,"Error: Invalid -markov specification '%s'\n",markov_string);
      return -2;
    }
    /* Check that probabilities make sense */
    if ((fabs(global.chain.matrix[0][0] + global.chain.matrix[0][1] + global.chain.matrix[0][2] - 1.0) > EPSILON) || 
        (fabs(global.chain.matrix[1][0] + global.chain.matrix[1][1] + global.chain.matrix[1][2] - 1.0) > EPSILON) || 
        (fabs(global.chain.matrix[2][0] + global.chain.matrix[2][1] + global.chain.matrix[2][2] - 1.0) > EPSILON)) {
      fprintf(stderr,"Error: Invalid Markov matrix\n");
      return -2;
    }
  } else {
    fprintf(stderr,"Error: At least one -markov flag is required\n");
    return -1;
  }

  if (SwitchPresent(pa,"-pareto") && SwitchPresent(pa,"-weibull")) {
    fprintf(stderr,"Only one of -pareto and -weibull flags allowed\n");
    return -1;
  }

  if (SwitchPresent(pa,"-pareto")) {
    global.distribution = PARETO;
    const char *pareto_string=SwitchValueN(pa,"-pareto",1,"0");
    if (*SwitchValueN(pa,"-length",2,""))
      fprintf(stderr,"Warning: multiple -pareto flags, ignoring all but the first one\n");
    sscanf(pareto_string,"%lf",&(global.distribution_shape));
    if (global.distribution_shape <= 1.0) {
      fprintf(stderr,"Error: Pareto parameter must be >1\n");
      return -1;
    }
  } else if (SwitchPresent(pa,"-weibull")) {
    global.distribution = WEIBULL;
    const char *weibull_string=SwitchValueN(pa,"-weibull",1,"0");
    if (*SwitchValueN(pa,"-length",2,""))
      fprintf(stderr,"Warning: multiple -weibull flags, ignoring all but the first one\n");
    sscanf(weibull_string,"%lf",&(global.distribution_shape));
    if (global.distribution_shape <= 0.0) {
      fprintf(stderr,"Error: Weibull parameter must be >0\n");
      return -1;
    }
  } else { /* markov */
    global.distribution = MARKOV;
    global.distribution_shape = -1.0;
  }

  /* -length switch */
  if (SwitchPresent(pa,"-length")) {
    const char *length_string=SwitchValueN(pa,"-length",1,"0");
    if (*SwitchValueN(pa,"-length",2,""))
      fprintf(stderr,"Warning: multiple -length flags, ignoring all but the first one\n");
    sscanf(length_string,"%d",&(global.length));
  }  else {
    fprintf(stderr,"Using default length: 100\n");
    global.length = 100;
  }

  return 0;
}

